from collections.abc import Sequence
from numbers import Integral
from typing import Optional, Union

import numpy as np
from rdkit.Chem import Mol
from scipy.sparse import csr_array
from sklearn.utils._param_validation import Interval

from skfp.bases import BaseFingerprintTransformer
from skfp.utils import ensure_mols


class ECFPFingerprint(BaseFingerprintTransformer):
    """
    Extended Connectivity FingerPrint (ECFP).

    The implementation uses RDKit. This is a hashed fingerprint [1]_, where
    fragments are computed based on circular substructures around each atom.
    Also known as Morgan fingerprint.

    Subgraphs are created around each atom with increasing radius, starting
    with just an atom itself. In each iteration, it is increased by another
    atom (one "hop" on the graph). Each subgraph during iteration is hashed,
    and the resulting hashes are folded to the ``fp_size`` length.

    Each subgraph gets an identifier based on atom types in its radius, which is
    then hashed. Atom types (invariants) by default are based on Daylight invariants:

    - number of heavy neighbors
    - valence (excluding hydrogen neighbors)
    - atomic number
    - atomic mass
    - formal charge
    - number of bound hydrogens
    - whether it is a part of a ring

    Alternatively, pharmacophoric invariants can be used, representing functional
    class of atoms, resulting in FCFP (Functional Circular FingerPrint) fingerprints.

    Parameters
    ----------
    fp_size : int, default=2048
        Size of output vectors, i.e. number of bits for each fingerprint. Must be
        positive.

    radius : int, default=2
        Number of iterations performed, i.e. maximum radius of resulting subgraphs.
        Another common notation uses diameter, therefore ECFP4 has radius 2.

    use_pharmacophoric_invariants : bool, default=False
        Whether to use pharmacophoric invariants (atom types) instead of default ones.
        They are: Donor, Acceptor, Aromatic, Halogen, Basic, Acidic. This results in
        FCFP (Functional Connectivity FingerPrint) fingerprint.

    include_chirality : bool, default=False
        Whether to include chirality information when computing atom types.

    use_bond_types : bool, default=True
        Whether to use bond types (single, double, triple, aromatic) when computing
        subgraphs.

    include_ring_membership : bool, default=True
        Whether to check if atom is part of a ring when computing atom types.

    count_simulation : bool, default=False
        Whether to use count simulation for approximating feature counts [2]_.

    count : bool, default=False
        Whether to return binary (bit) features, or their counts.

    sparse : bool, default=False
        Whether to return dense NumPy array, or sparse SciPy CSR array.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized
        over the input molecules. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when computing fingerprints.

    Attributes
    ----------
    n_features_out : int
        Number of output features. Equal to ``fp_size``.

    requires_conformers : bool = False
        This fingerprint uses only 2D molecular graphs and does not require conformers.

    See Also
    --------
    :class:`SECFPFingerprint` : Related fingerprint, which additionally uses shortest paths
        between atoms like Atom Pair fingerprint.

    :class:`MHFPFingerprint` : Related fingerprint, which uses MinHash to perform hashing
        and can compute raw hashes, instead of folded vectors.

    References
    ----------
    .. [1] `David Rogers and Mathew Hahn
        "Extended-Connectivity Fingerprints"
        J. Chem. Inf. Model. 2010, 50, 5, 742–754
        <https://pubs.acs.org/doi/10.1021/ci100050t>`_

    .. [2] `Gregory Landrum
        "Simulating count fingerprints"
        <https://greglandrum.github.io/rdkit-blog/posts/2021-07-06-simulating-counts.html>`_

    Examples
    --------
    >>> from skfp.fingerprints import ECFPFingerprint
    >>> smiles = ["O", "CC", "[C-]#N", "CC=O"]
    >>> fp = ECFPFingerprint()
    >>> fp
    ECFPFingerprint()

    >>> fp.transform(smiles)
    array([[0, 0, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 0, 0, 0]], dtype=uint8)
    """

    _parameter_constraints: dict = {
        **BaseFingerprintTransformer._parameter_constraints,
        "fp_size": [Interval(Integral, 1, None, closed="left")],
        "radius": [Interval(Integral, 0, None, closed="left")],
        "use_pharmacophoric_invariants": ["boolean"],
        "include_chirality": ["boolean"],
        "use_bond_types": ["boolean"],
        "include_ring_membership": ["boolean"],
        "use_2D": ["boolean"],
    }

    def __init__(
        self,
        fp_size: int = 2048,
        radius: int = 2,
        use_pharmacophoric_invariants: bool = False,
        include_chirality: bool = False,
        use_bond_types: bool = True,
        include_ring_membership: bool = True,
        count_simulation: bool = False,
        count: bool = False,
        sparse: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            n_features_out=fp_size,
            count=count,
            sparse=sparse,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )
        self.fp_size = fp_size
        self.radius = radius
        self.use_pharmacophoric_invariants = use_pharmacophoric_invariants
        self.include_chirality = include_chirality
        self.use_bond_types = use_bond_types
        self.include_ring_membership = include_ring_membership
        self.count_simulation = count_simulation

    def _calculate_fingerprint(
        self, X: Sequence[Union[str, Mol]]
    ) -> Union[np.ndarray, csr_array]:
        from rdkit.Chem.rdFingerprintGenerator import (
            GetMorganFeatureAtomInvGen,
            GetMorganGenerator,
        )

        X = ensure_mols(X)

        if self.use_pharmacophoric_invariants:
            inv_gen = GetMorganFeatureAtomInvGen()
        else:
            inv_gen = None

        gen = GetMorganGenerator(
            fpSize=self.fp_size,
            radius=self.radius,
            atomInvariantsGenerator=inv_gen,
            includeChirality=self.include_chirality,
            useBondTypes=self.use_bond_types,
            includeRingMembership=self.include_ring_membership,
            countSimulation=self.count_simulation,
        )

        if self.count:
            X = [gen.GetCountFingerprintAsNumPy(mol) for mol in X]
        else:
            X = [gen.GetFingerprintAsNumPy(mol) for mol in X]

        return csr_array(X) if self.sparse else np.array(X)



from typing import Union

import numpy as np
from sklearn.metrics import roc_auc_score
from sklearn.utils._param_validation import (
    Interval,
    RealNotInt,
    StrOptions,
    validate_params,
)


@validate_params(
    {
        "y_true": ["array-like"],
        "y_pred": ["array-like"],
        "constant_target_behavior": [
            StrOptions({"raise"}),
            "nan",
            None,
            Interval(RealNotInt, 0, 1, closed="both"),
        ],
    },
    prefer_skip_nested_validation=True,
)
def auroc_score(
    y_true: Union[np.ndarray, list[float]],
    y_score: Union[np.ndarray, list[float]],
    *args,
    constant_target_behavior: Union[str, float] = np.NaN,
    **kwargs,
) -> float:
    """
    Area Under Receiver Operating Characteristic curve (AUROC / ROC AUC).

    Wrapper around scikit-learn ``roc_auc_score`` function, which adds
    ``constant_target_behavior`` to control behavior for all-zero ``y_true`` labels.
    Scikit-learn behavior is to throw an error, since AUROC is undefined there, but
    this can easily happen for cross-validation in imbalanced problems.

    Parameters
    ----------
    y_true : array-like of shape (n_samples,) or (n_samples, n_outputs)
        Ground truth (correct) target values.

    y_score : array-like of shape (n_samples,) or (n_samples, n_outputs)
        Target scores, i.e. probability of the class with the greater label for each
        output** of the classifier.

    *args, **kwargs
        Any additional parameters for the underlying ``roc_auc_score`` function.

    constant_target_behavior : "raise", np.NaN, None, or float, default=np.NaN
        Value returned if ``y_true`` contains only constant values. ``"raise"`` is the
        default scikit-learn behavior, which raises an error. Default ``np.NaN``
        (or None) ignores the given fold in cross-validation. Alternatively, float value
        (e.g. 0.5, 1.0) can be returned.

    Returns
    -------
    score : float
        AUROC value.

    Examples
    --------
    >>> import numpy as np
    >>> from skfp.metrics import auroc_score
    >>> y_true = np.array([0, 0, 0])
    >>> y_score = np.array([0.5, 0.6, 0.7])
    >>> auroc_score(y_true, y_score)
    nan
    >>> auroc_score(y_true, y_score, constant_target_behavior=0.5)
    0.5
    """

    try:
        return roc_auc_score(y_true, y_score, *args, **kwargs)
    except ValueError as e:
        if (
            "Only one class present in y_true" in str(e)
            and constant_target_behavior != "raise"
        ):
            return constant_target_behavior  # type: ignore
        else:
            raise



from typing import Optional

from rdkit.Chem import Mol
from rdkit.Chem.Crippen import MolLogP
from rdkit.Chem.Descriptors import MolWt
from rdkit.Chem.rdMolDescriptors import CalcNumHBA, CalcNumHBD

from skfp.bases.base_filter import BaseFilter


class RuleOfTwoFilter(BaseFilter):
    """
    Rule of two (Ro2).

    Designed for finding reagents for building block design [1]_.

    Molecule must fulfill conditions:

    - molecular weight <= 200 daltons
    - HBA <= 4
    - HBD <= 2
    - logP <= 2

    Parameters
    ----------
    allow_one_violation : bool, default=False
        Whether to allow violating one of the rules for a molecule. This makes the
        filter less restrictive.

    return_indicators : bool, default=False
        Whether to return a binary vector with indicators which molecules pass the
        filter, instead of list of molecules.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform_x_y` and
        :meth:`transform` are parallelized over the input molecules. ``None`` means 1
        unless in a :obj:`joblib.parallel_backend` context. ``-1`` means using all
        processors. See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when filtering molecules.

    References
    -----------
    .. [1] `Goldberg, F. W., Kettle, J. G., Kogej, T., Perry, M. W. D., & Tomkinson, N. P.
        "Designing novel building blocks is an overlooked strategy to improve compound quality."
        Drug Discovery Today, 20(1), 11–17.
        <https://doi.org/10.1016/j.drudis.2014.09.023>`_

    Examples
    ----------
    >>> from skfp.filters import RuleOfTwoFilter
    >>> smiles = ['C=CCc1c(C)[nH]c(N)nc1=O', 'C=CCNC(=O)c1ccncc1', 'C=CCC1C=C(C)CC(CC=C)N1']
    >>> filt = RuleOfTwoFilter()
    >>> filt
    RuleOfTwoFilter()
    >>> filtered_mols = filt.transform(smiles)
    >>> filtered_mols
    ['C=CCc1c(C)[nH]c(N)nc1=O', 'C=CCNC(=O)c1ccncc1']
    """

    def __init__(
        self,
        allow_one_violation: bool = False,
        return_indicators: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            allow_one_violation=allow_one_violation,
            return_indicators=return_indicators,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )

    def _apply_mol_filter(self, mol: Mol) -> bool:
        rules = [
            MolWt(mol) <= 200,
            CalcNumHBA(mol) <= 4,
            CalcNumHBD(mol) <= 2,
            MolLogP(mol) <= 2,
        ]
        passed_rules = sum(rules)

        if self.allow_one_violation:
            return passed_rules >= len(rules) - 1
        else:
            return passed_rules == len(rules)



from numbers import Integral
from typing import Optional

from rdkit.Chem import Mol
from rdkit.Chem.Descriptors import MolWt
from sklearn.utils._param_validation import Interval, InvalidParameterError

from skfp.bases.base_filter import BaseFilter


class MolecularWeightFilter(BaseFilter):
    """
    Molecular weight filter.

    Filters out molecules with mass in Daltons outside the given range. Provided
    `min_weight` and `max_weight` are inclusive on both sides.

    Parameters
    ----------
    min_weight : int, default=0
        Minimal weight in Daltons.

    max_weight : int, default=1000
        Maximal weight in Daltons.

    return_indicators : bool, default=False
        Whether to return a binary vector with indicators which molecules pass the
        filter, instead of list of molecules.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform_x_y` and
        :meth:`transform` are parallelized over the input molecules. ``None`` means 1
        unless in a :obj:`joblib.parallel_backend` context. ``-1`` means using all
        processors. See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when filtering molecules.

    Examples
    --------
    >>> from skfp.filters import MolecularWeightFilter
    >>> smiles = ["C", "O", "CN1C=NC2=C1C(=O)N(C(=O)N2C)C"]
    >>> filt = MolecularWeightFilter(max_weight=100)
    >>> filt
    MolecularWeightFilter(max_weight=100)

    >>> filtered_mols = filt.transform(smiles)
    >>> filtered_mols
    ['C', 'O']
    """

    _parameter_constraints: dict = {
        **BaseFilter._parameter_constraints,
        "min_weight": [Interval(Integral, 0, None, closed="left")],
        "max_weight": [Interval(Integral, 0, None, closed="left")],
    }

    def __init__(
        self,
        min_weight: int = 0,
        max_weight: int = 1000,
        return_indicators: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            return_indicators=return_indicators,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )
        self.min_weight = min_weight
        self.max_weight = max_weight

    def _validate_params(self) -> None:
        super()._validate_params()
        if self.max_weight < self.min_weight:
            raise InvalidParameterError(
                f"The max_weight parameter of {self.__class__.__name__} must be "
                f"greater or equal to min_weight, got: "
                f"min_weight={self.min_weight}, max_weight={self.max_weight}"
            )

    def _apply_mol_filter(self, mol: Mol) -> bool:
        return self.min_weight <= MolWt(mol) <= self.max_weight



from abc import ABC, abstractmethod
from collections.abc import Sequence
from copy import deepcopy
from numbers import Integral
from typing import Any, Optional, Union

import numpy as np
import scipy.sparse
from joblib import effective_n_jobs
from rdkit.Chem.rdchem import Mol
from rdkit.DataStructs import (
    IntSparseIntVect,
    LongSparseIntVect,
    SparseBitVect,
    UIntSparseIntVect,
    ULongSparseIntVect,
)
from scipy.sparse import csr_array, dok_array
from sklearn.base import (
    BaseEstimator,
    ClassNamePrefixFeaturesOutMixin,
    TransformerMixin,
)
from sklearn.utils._param_validation import InvalidParameterError
from tqdm import tqdm

from skfp.utils import run_in_parallel

"""
If you get MaybeEncodingError, first check any worker functions for exceptions!
That error isn't very informative, but gets thrown in Joblib multiprocessing.
"""

"""
Note that you need to do create RDKit objects *inside* the function that runs
in parallel, i.e. _calculate_fingerprint(), not in the constructor or outside in
general.
This is because Joblib needs to pickle data sent to workers, and RDKit objects
cannot be pickled, throwing TypeError: cannot pickle 'Boost.Python.function' object
"""


class BaseFingerprintTransformer(
    ABC, BaseEstimator, TransformerMixin, ClassNamePrefixFeaturesOutMixin
):
    """Base class for fingerprint transformers."""

    # parameters common for all fingerprints
    _parameter_constraints: dict = {
        "count": ["boolean"],
        "sparse": ["boolean"],
        "n_jobs": [Integral, None],
        "batch_size": [Integral, None],
        "verbose": ["verbose"],
        "random_state": ["random_state"],
    }

    def __init__(
        self,
        n_features_out: int,
        requires_conformers: bool = False,
        count: bool = False,
        sparse: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
        random_state: Optional[int] = 0,
    ):
        self.count = count
        self.sparse = sparse
        self.n_jobs = n_jobs
        self.batch_size = batch_size
        self.verbose = verbose
        self.random_state = random_state

        # this, combined with ClassNamePrefixFeaturesOutMixin, automatically handles
        # set_output() API
        self._n_features_out = n_features_out
        self.n_features_out = self._n_features_out

        # indicate whether inputs need to be molecules with conformers computed and
        # conf_id integer property set; this allows programmatically checking which
        # fingerprints are 3D-based and require such input
        self.requires_conformers = requires_conformers

    def __sklearn_is_fitted__(self) -> bool:
        return True  # fingerprint transformers don't need fitting

    def set_params(self, **params):
        super().set_params(**params)
        # for fingerprints that have both 2D and 3D versions, as indicated by use_3D
        # attribute, we need to also keep requires_conformers attribute in sync
        if hasattr(self, "use_3D"):
            self.requires_conformers = self.use_3D
        return self

    def fit(self, X: Sequence[Union[str, Mol]], y: Optional[Any] = None, **fit_params):
        """
        Unused, kept for Scikit-learn compatibility.

        Parameters
        ----------
        X : any
            Unused, kept for Scikit-learn compatibility.

        y : any
            Unused, kept for Scikit-learn compatibility.

        **fit_params : dict
            Unused, kept for Scikit-learn compatibility.

        Returns
        --------
        self
        """
        self._validate_params()
        return self

    def fit_transform(
        self, X: Sequence[Union[str, Mol]], y: Optional[Any] = None, **fit_params
    ):
        """
        The same as ``.transform()`` method, kept for Scikit-learn compatibility.

        Parameters
        ----------
        X : any
            See ``.transform()`` method.

        y : any
            See ``.transform()`` method.

        **fit_params : dict
            Unused, kept for Scikit-learn compatibility.

        Returns
        -------
        X_new : any
            See ``.transform()`` method.
        """
        return self.transform(X)

    def transform(
        self, X: Sequence[Union[str, Mol]], copy: bool = False
    ) -> Union[np.ndarray, csr_array]:
        self._validate_params()

        if copy:
            X = deepcopy(X)

        n_jobs = effective_n_jobs(self.n_jobs)
        if n_jobs == 1:
            if self.verbose:
                results = [self._calculate_fingerprint([mol]) for mol in tqdm(X)]
            else:
                results = self._calculate_fingerprint(X)
        else:
            results = run_in_parallel(
                self._calculate_fingerprint,
                data=X,
                n_jobs=n_jobs,
                batch_size=self.batch_size,
                verbose=self.verbose,
            )

        if isinstance(results, (np.ndarray, csr_array)):
            return results
        else:
            return scipy.sparse.vstack(results) if self.sparse else np.vstack(results)

    @abstractmethod
    def _calculate_fingerprint(
        self, X: Sequence[Union[str, Mol]]
    ) -> Union[np.ndarray, csr_array]:
        """
        Calculate fingerprints for a given input batch.

        :param X: subset of original X data
        :return: array containing calculated fingerprints for each molecule
        """
        raise NotImplementedError

    def _validate_params(self) -> None:
        # override Scikit-learn validation to make stacktrace nicer
        try:
            super()._validate_params()
        except InvalidParameterError as e:
            raise InvalidParameterError(str(e)) from None

    @staticmethod
    def _hash_fingerprint_bits(
        X: list[
            Union[
                IntSparseIntVect,
                LongSparseIntVect,
                SparseBitVect,
                UIntSparseIntVect,
                ULongSparseIntVect,
            ]
        ],
        fp_size: int,
        count: bool,
        sparse: bool,
    ) -> Union[np.ndarray, csr_array]:
        rdkit_vec_types = (
            IntSparseIntVect,
            LongSparseIntVect,
            SparseBitVect,
            UIntSparseIntVect,
            ULongSparseIntVect,
        )
        if not all(isinstance(x, rdkit_vec_types) for x in X):
            raise ValueError(
                f"Fingerprint hashing requires instances of one of: {rdkit_vec_types}"
            )

        shape = (len(X), fp_size)
        dtype = np.uint32 if count else np.uint8
        arr = dok_array(shape, dtype=dtype) if sparse else np.zeros(shape, dtype=dtype)

        if isinstance(X[0], SparseBitVect):
            for idx, x in enumerate(X):
                for fp_bit in x.GetOnBits():
                    arr[idx, fp_bit % fp_size] += 1
        else:
            for idx, x in enumerate(X):
                for fp_bit, val in x.GetNonzeroElements().items():
                    arr[idx, fp_bit % fp_size] += val

        arr = arr.tocsr() if sparse else arr
        arr = (arr > 0).astype(np.uint8) if not count else arr

        return arr



import os
from typing import Optional, Union

import numpy as np
import pandas as pd
from sklearn.utils._param_validation import validate_params

from skfp.datasets.utils import fetch_dataset, get_mol_strings_and_labels


@validate_params(
    {
        "data_dir": [None, str, os.PathLike],
        "as_frame": ["boolean"],
        "verbose": ["boolean"],
    },
    prefer_skip_nested_validation=True,
)
def load_freesolv(
    data_dir: Optional[Union[str, os.PathLike]] = None,
    as_frame: bool = False,
    verbose: bool = False,
) -> Union[pd.DataFrame, tuple[list[str]], np.ndarray]:
    """
    Load and return the FreeSolv (Free Solvation Database) dataset.

    The task is to predict hydration free energy of small molecules in water [1]_ [2]_.
    Targets are in kcal/mol.

    ==================   ==============
    Tasks                             1
    Task type                regression
    Total samples                   642
    Recommended split          scaffold
    Recommended metric             RMSE
    ==================   ==============

    Parameters
    ----------
    data_dir : {None, str, path-like}, default=None
        Path to the root data directory. If ``None``, currently set scikit-learn directory
        is used, by default `$HOME/scikit_learn_data`.

    as_frame : bool, default=False
        If True, returns the raw DataFrame with columns: "SMILES", "label". Otherwise,
        returns SMILES as list of strings, and labels as a NumPy array (1D float vector).

    verbose : bool, default=False
        If True, progress bar will be shown for downloading or loading files.

    Returns
    -------
    data : pd.DataFrame or tuple(list[str], np.ndarray)
        Depending on the ``as_frame`` argument, one of:
        - Pandas DataFrame with columns: "SMILES", "label"
        - tuple of: list of strings (SMILES), NumPy array (labels)

    References
    ----------
    .. [1] `Mobley, D.L., Guthrie, J.P.
        "FreeSolv: a database of experimental and calculated hydration free energies,
        with input files"
        J Comput Aided Mol Des 28, 711–720 (2014)
        <https://link.springer.com/article/10.1007/s10822-014-9747-x>`_

    .. [2] `Zhenqin Wu et al.
        "MoleculeNet: a benchmark for molecular machine learning"
        Chem. Sci., 2018,9, 513-530
        <https://pubs.rsc.org/en/content/articlelanding/2018/sc/c7sc02664a>`_
    """
    df = fetch_dataset(
        data_dir,
        dataset_name="MoleculeNet_FreeSolv",
        filename="freesolv.csv",
        verbose=verbose,
    )
    return df if as_frame else get_mol_strings_and_labels(df)



from collections.abc import Sequence
from typing import Optional, Union

import numpy as np
from mordred import Calculator, descriptors
from rdkit.Chem import Mol
from scipy.sparse import csr_array

from skfp.bases import BaseFingerprintTransformer
from skfp.utils import ensure_mols


class MordredFingerprint(BaseFingerprintTransformer):
    """
    Mordred fingerprint.

    The implementation uses ``mordredcommunity`` [3]_ library. This is a descriptor-based
    fingerprint, implementing a very large number of 2D, and optionally 3D, molecular
    descriptors, originally implemented in the Mordred [2]_ library.

    Descriptors include simple counts (e.g. atom types, rings), topological indices,
    computed properties (e.g. ClogP, polarizability), and more. For a full list, see
    Supplementary File 3 in the original publication [1]_.

    Parameters
    ----------
    use_3D : bool, default=False
        Whether to include 3D (conformer-based) descriptors. Using this option results
        in longer computation time.

    sparse : bool, default=False
        Whether to return dense NumPy array, or sparse SciPy CSR array.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized
        over the input molecules. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when computing fingerprints.

    Attributes
    ----------
    n_features_out : int
        Number of output features, size of fingerprints: 1613 for 2D-only variant,
        and 1826 if using 3D descriptors.

    requires_conformers : bool = False
        This fingerprint in the 3D variant computes conformers internally, and
        therefore does not require conformers.

    References
    ----------
    .. [1] `Moriwaki, Hirotomo, et al.
        "Mordred: a molecular descriptor calculator"
        J Cheminform 10, 4 (2018)
        <https://jcheminf.biomedcentral.com/articles/10.1186/s13321-018-0258-y>`_

    .. [2] https://github.com/mordred-descriptor/mordred

    .. [3] https://github.com/JacksonBurns/mordred-community

    Examples
    --------
    >>> from skfp.fingerprints import MordredFingerprint
    >>> smiles = ["O", "CC", "[C-]#N", "CC=O"]
    >>> fp = MordredFingerprint()
    >>> fp
    MordredFingerprint()

    >>> fp.transform(smiles)
    array([[0.       , 0.       , 0.       , ..., 0.       ,       nan,
            0.       ],
           [0.       , 0.       , 0.       , ..., 1.       , 2.       ,
            1.       ],
           [0.       , 0.       , 1.       , ..., 1.       , 2.       ,
            1.       ],
           [1.4142135, 1.4142135, 0.       , ..., 4.       , 2.25     ,
            1.       ]], dtype=float32)
    """

    _parameter_constraints: dict = {
        **BaseFingerprintTransformer._parameter_constraints,
        "use_3D": ["boolean"],
    }

    def __init__(
        self,
        use_3D: bool = False,
        sparse: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        n_features_out = 1826 if use_3D else 1613
        super().__init__(
            n_features_out=n_features_out,
            requires_conformers=use_3D,
            sparse=sparse,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )
        self.use_3D = use_3D

    def _calculate_fingerprint(
        self, X: Sequence[Union[str, Mol]]
    ) -> Union[np.ndarray, csr_array]:
        X = ensure_mols(X)

        calc = Calculator(descriptors, ignore_3D=not self.use_3D)
        X = [calc(mol) for mol in X]

        return (
            csr_array(X, dtype=np.float32)
            if self.sparse
            else np.array(X, dtype=np.float32)
        )



from .bace import load_bace
from .bbbp import load_bbbp
from .benchmark import load_moleculenet_benchmark, load_ogb_splits
from .clintox import load_clintox
from .esol import load_esol
from .freesolv import load_freesolv
from .hiv import load_hiv
from .lipophilicity import load_lipophilicity
from .muv import load_muv
from .pcba import load_pcba
from .sider import load_sider
from .tox21 import load_tox21
from .toxcast import load_toxcast



from typing import Optional

from rdkit.Chem import FilterCatalog, Mol
from rdkit.Chem.rdfiltercatalog import FilterCatalogParams

from skfp.bases.base_filter import BaseFilter


class GlaxoFilter(BaseFilter):
    """
    Glaxo filter.

    Designed at Glaxo Wellcome (currently GSK) to filter out molecules with reactive
    functional groups, unsuitable leads (i.e. compounds which would not be initially
    followed up), and unsuitable natural products (i.e., derivatives of natural product
    compounds known to interfere with to interfere with common assay procedures).

    Rule definitions are available in the supplementary material of the original
    publication [1]_ and in RDKit code [2]_.

    Parameters
    ----------
    allow_one_violation : bool, default=False
        Whether to allow violating one of the rules for a molecule. This makes the
        filter less restrictive.

    return_indicators : bool, default=False
        Whether to return a binary vector with indicators which molecules pass the
        filter, instead of list of molecules.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform_x_y` and
        :meth:`transform` are parallelized over the input molecules. ``None`` means 1
        unless in a :obj:`joblib.parallel_backend` context. ``-1`` means using all
        processors. See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when filtering molecules.

    References
    ----------
    .. [1] `Mike Hann et al.
        "Strategic Pooling of Compounds for High-Throughput Screening"
        J. Chem. Inf. Comput. Sci. 1999, 39, 5, 897–902
        <https://pubs.acs.org/doi/10.1021/ci990423o>`_

    .. [2] `RDKit Glaxo filter definitions
        <https://github.com/rdkit/rdkit/blob/e4f4644a89d6446ddebda0bf396fa4335324c41c/Code/GraphMol/FilterCatalog/chembl_glaxo.in>`_

    Examples
    --------
    >>> from skfp.filters import GlaxoFilter
    >>> smiles = ["C", "O", "N=C=N"]
    >>> filt = GlaxoFilter()
    >>> filt
    GlaxoFilter()

    >>> filtered_mols = filt.transform(smiles)
    >>> filtered_mols
    ['C', 'O']
    """

    def __init__(
        self,
        allow_one_violation: bool = False,
        return_indicators: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            allow_one_violation=allow_one_violation,
            return_indicators=return_indicators,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )
        self._filters = self._load_filters()

    def _load_filters(self) -> FilterCatalog:
        filter_rules = FilterCatalogParams.FilterCatalogs.CHEMBL_Glaxo
        params = FilterCatalog.FilterCatalogParams()
        params.AddCatalog(filter_rules)
        filters = FilterCatalog.FilterCatalog(params)
        return filters

    def _apply_mol_filter(self, mol: Mol) -> bool:
        errors = len(self._filters.GetMatches(mol))
        return not errors or (self.allow_one_violation and errors == 1)



import os
from typing import Optional, Union

import numpy as np
import pandas as pd
from sklearn.utils._param_validation import StrOptions, validate_params

from skfp.datasets.utils import fetch_dataset, get_mol_strings_and_labels


@validate_params(
    {
        "data_dir": [None, str, os.PathLike],
        "mol_type": [StrOptions({"SMILES", "aminoseq"})],
        "as_frame": ["boolean"],
        "verbose": ["boolean"],
    },
    prefer_skip_nested_validation=True,
)
def load_peptides_func(
    data_dir: Optional[Union[str, os.PathLike]] = None,
    mol_type: str = "SMILES",
    as_frame: bool = False,
    verbose: bool = False,
) -> Union[pd.DataFrame, tuple[list[str]], np.ndarray]:
    """
    Load and return the Peptides-func dataset.

    The task is to predict functional properties (binary classes) for a set
    of peptides (small proteins) [1]_.

    ==================   =================
    Tasks                               10
    Task type               classification
    Total samples                    15535
    Recommended split    stratified random
    Recommended metric               AUPRC
    ==================   =================

    Parameters
    ----------
    data_dir : {None, str, path-like}, default=None
        Path to the root data directory. If ``None``, currently set scikit-learn directory
        is used, by default `$HOME/scikit_learn_data`.

    mol_type : {"SMILES", "aminoseq"}, default="SMILES"
        Which molecule representation to return, either SMILES strings or aminoacid
        sequences.

    as_frame : bool, default=False
        If True, returns the raw DataFrame with columns: "SMILES"/"aminoseq" and "label".
        This depends on the ``mol_type`` parameter. Otherwise, returns molecules as
        a list of strings (either SMILES or aminoacid sequences), and labels as a NumPy
        array (2D integer array).

    verbose : bool, default=False
        If True, progress bar will be shown for downloading or loading files.

    Returns
    -------
    data : pd.DataFrame or tuple(list[str], np.ndarray)
        Depending on the ``as_frame`` and ``mol_type`` parameters, one of:
        - Pandas DataFrame with columns: "SMILES"/"aminoseq", "label"
        - tuple of: list of strings (SMILES / aminoacid sequences), NumPy array (labels)

    References
    ----------
    .. [1] `Dwivedi, Vijay Prakash, et al.
        "Long Range Graph Benchmark"
        Advances in Neural Information Processing Systems 35 (2022): 22326-22340
        <https://proceedings.neurips.cc/paper_files/paper/2022/hash/8c3c666820ea055a77726d66fc7d447f-Abstract-Datasets_and_Benchmarks.html>`_
    """
    df = fetch_dataset(
        data_dir,
        dataset_name="LRGB_Peptides-func",
        filename="peptides_func.csv",
        verbose=verbose,
    )
    return df if as_frame else get_mol_strings_and_labels(df, mol_type=mol_type)



import functools
from collections.abc import Iterable

from rdkit.Chem import Atom, Mol, MolFromSmarts, rdMolDescriptors


def get_num_carbon_atoms(mol: Mol) -> int:
    """
    Calculated number of carbon atoms in a molecule.
    """
    return sum(atom.GetSymbol() == "C" for atom in mol.GetAtoms())


def get_num_heavy_metals(mol: Mol) -> int:
    """
    Calculated number of heavy metals in a molecule.

    Heavy atoms are defined as metals other than ["Li", "Be", "K", "Na", "Ca", "Mg"].
    """
    # non-metals and non-heavy metals
    # fmt: off
    not_heavy_metals = {
        "He", "Li", "Be", "B", "C", "N", "O", "F", "Ne", "Na", "Mg", "Si", "P", "S",
        "Cl", "Ar", "K", "Ca", "As", "Se", "Br", "Kr", "Te", "I", "Xe", "At", "Rn"
    }
    # fmt: on

    # easier to reject than to list all heavy metals
    return sum(atom.GetSymbol() not in not_heavy_metals for atom in mol.GetAtoms())


def get_num_charged_atoms(mol: Mol) -> int:
    """
    Calculated number of charged atoms in a molecule.
    """
    return sum(atom.GetFormalCharge() != 0 for atom in mol.GetAtoms())


def get_num_rigid_bonds(mol: Mol) -> int:
    """
    Calculates number of rigid bonds in a molecule.
    """
    total_bonds = mol.GetNumBonds()
    rotatable_bonds = rdMolDescriptors.CalcNumRotatableBonds(mol)
    return total_bonds - rotatable_bonds


def get_num_aromatic_rings(mol: Mol) -> int:
    """
    Calculates number of aromatic rings in a molecule.
    """
    return sum(is_ring_aromatic(mol, ring) for ring in mol.GetRingInfo().AtomRings())


def get_max_num_fused_aromatic_rings(mol: Mol) -> int:
    """
    Calculates the number of rings in the largest system of fused
    aromatic rings in a molecule.
    """
    ring_info = mol.GetRingInfo()

    fused_aromatic_ring_idxs = []
    for idx, ring in enumerate(mol.GetRingInfo().AtomRings()):
        if is_ring_aromatic(mol, ring) and ring_info.IsRingFused(idx):
            fused_aromatic_ring_idxs.append(idx)

    # find the largest fused ring system by exhaustively checking
    # all combinations

    largest_size = 0
    for i, ring_idx in enumerate(fused_aromatic_ring_idxs):
        fused_rings = [
            ring_idx_2
            for ring_idx_2 in fused_aromatic_ring_idxs[i + 1 :]
            if ring_info.AreRingsFused(ring_idx, ring_idx_2)
        ]
        if fused_rings:
            # original ring and fused ones
            largest_size = max(largest_size, 1 + len(fused_rings))

    return largest_size


def is_ring_aromatic(mol: Mol, ring_atoms: Iterable[Atom]) -> bool:
    """
    Checks whether a ring is aromatic.
    """
    for atom_idx in ring_atoms:
        if not mol.GetAtomWithIdx(atom_idx).GetIsAromatic():
            return False

    return True


def get_max_ring_size(mol: Mol) -> int:
    """
    Calculates maximum ring size in a molecule.
    """
    rings = mol.GetRingInfo().AtomRings()
    return max(len(ring) for ring in rings) if rings else 0


def get_non_carbon_to_carbon_ratio(mol: Mol) -> float:
    """
    Calculates ratio of non-carbon to carbon atoms.
    """
    num_carbons = get_num_carbon_atoms(mol)
    num_non_carbons = mol.GetNumAtoms() - num_carbons
    return num_non_carbons / num_carbons if num_carbons > 0 else 0.0


def get_num_charged_functional_groups(mol: Mol) -> int:
    """
    Calculates the number of charged functional groups in a molecule, i.e. having
    a non-zero total charge.

    Since determining what is a functional group is an arguable topic with no
    definite answers, we use a list of CXSMARTS by ChemAxon:
    https://docs.chemaxon.com/display/docs/attachments/attachments_1829721_1_functionalgroups.cxsmi.

    Phosphine and sulfoxide patterns could not be parsed by RDKit, so we manually fixed
    them.
    """
    atomic_charges = {atom.GetIdx(): atom.GetFormalCharge() for atom in mol.GetAtoms()}
    fragment_smarts = _get_functional_groups_smarts_patterns()

    matched_groups: list[tuple[int]] = []
    for smarts in fragment_smarts:
        # note that this returns lists of atom indexes
        matches = mol.GetSubstructMatches(smarts)
        matches = list(matches) if matches else []
        matched_groups.extend(matches)

    num_charged_groups = sum(
        sum(atomic_charges[atom_idx] for atom_idx in group) != 0
        for group in matched_groups
    )

    return num_charged_groups


@functools.cache
def _get_functional_groups_smarts_patterns() -> list[Mol]:
    # using cached function compiles SMARTS patterns only once, on first usage
    # flake8: noqa: E501
    fragment_smarts = [
        "[O&X2:2][C&X4&!$(CO~[!#6]):1][O&X2:3]",  # acetal
        "[C&X3:1](=[O&X1:2])[F,Cl,Br,I:3]",  # acid halide
        "[C:1][O&H1&!$(OC~[!#1&!#6]):2]",  # alcohol
        "[C&X3&H1,H2:1]=[O&X1&!$(O=C~[!#1&!#6]):2]",  # aldehyde
        "[C&H1:1]=[N:2][O&H1:3]",  # aldoxime
        "[A:1]",  # aliphatic
        "[C&X4:1][N&X3&!$(NC~[!#1&!#6&!F&!Cl&!Br&!I])&!$(N~[!#1&!#6])&!$(Nc):2]",  # aliphatic amine
        "[!#1&!#6&A:1]",  # aliphatic hetero
        "[#3,#11,#19,#37,#55,#87:1]",  # alkali metal
        "[#4,#12,#20,#38,#56,#88:1]",  # alkaline earth metal
        "[C:1]=[C:1]",  # alkene
        "[C&X2:1](=[C:2])=[C:2]",  # allene
        "[C&H2:3]=[C&H1:2][C&H2:1]",  # allyl
        "[C:1][O&-&!$(OC~[!#1&!#6]):2]",  # alkoxide
        "[C:1][O&X2&H0&!$(OC~[!#1&!#6]):2]",  # alkoxy
        "[C&X2:1]#[C&X2:1]",  # alkyne
        "[C&X4:1][F,Cl,Br,I;!$([F,Cl,Br,I]C~[!#1&!#6]):2]",  # alkyl halide
        "[N&X3:2][C&X3:1]=[O&X1:3]",  # amide
        "[N&X3:3][C:1]=[N&X2:2]",  # amidine
        "[#6:1][N&X3&!$(NC~[!#1&!#6&!F&!Cl&!Br&!I])&!$(N~[!#1&!#6]):2]",  # amine
        "[#6:1][N&X4&+:2]",  # ammonium
        "[C&X3:1](=[O&X1:2])[O&X2:3][C&X3:1]=[O&X1:2]",  # anhydride
        "[c:1][N&X3&!$(NC~[!#1&!#6])&!$(N~[!#1&!#6]):2]",  # aniline
        "[*:1]",  # any atom
        "[a:1]",  # aromatic
        "[!#1&!#6&a:1]",  # aromatic hetero
        "[c&X3:1][F,Cl,Br,I:2]",  # aryl halide
        "[C:1]1[C:1][N:2]1",  # aziridine
        "[#6:2]/[N&X2:1]=[N&X2:1]/[#6:2]",  # azo
        "[N&X2:2]=[N&X2:3][S&X4:1](=[O:5])=[O:4]",  # azosulfone
        "[#6:1][$([N&X2:2]=[N&X3&+:3]([O&-:4])[#6:5]),$([N&X2:2]=[N&X3&+0:3](=[O:4])[#6:5])]",  # azoxy
        "[C&H2]-c1[c&H1][c&H1][c&H1][c&H1][c&H1]1",  # benzyl
        "[Br&-:1]",  # bromide
        "[N&X3,N&X4&+:2][C&X3:1](=[O&X1:3])[O&X2,O&X1&-:4]",  # carbamate
        "[C:1](=[O&X1:2])([O:3][#6:4])[O:3][#6:4]",  # carbonate
        "[C&X3:1]=[O&X1:2]",  # oxo
        "[C&X3:1](=[O:2])[O&X2&H1:3]",  # carboxylic acid
        "[C&X3:1](=[O:2])[O&-:3]",  # carboxylate
        "[R0:1]",  # chain
        "[Cl&-:1]",  # chloride
        # CoA
        "[#1]OC([#1])(C(=O)N([#1])C([#1])([#1])C([#1])([#1])C(=O)N([#1])C([#1])([#1])[C:2]([#1])([#1])[S:1])C(C([#1])([#1])[#1])(C([#1])([#1])[#1])C([#1])([#1])OP(=O)(O[#1])OP(=O)(O[#1])OC([#1])([#1])C1([#1])OC([#1])(n2c([#1])nc3c(nc([#1])nc23)N([#1])[#1])C([#1])(O[#1])C1([#1])OP(=O)(O[#1])O[#1]",
        "[N&X3:3][C&X2:1]#[N&X1:2]",  # cyanamide
        "[N&X3:3][C&X2:1]#[N&X1:2]",  # cyanimide
        "[$([#6:1]=[N&+:2]=[N&-:3]),$([#6&-:1]-[N&+:2]#[N:3])]",  # diazo
        "[#6&+0:1]-[N&+:2]#[N:3]",  # diazonium
        "[#6:1][S&X2&H0:2][S&X2&H0:2][#6:1]",  # disulfide
        "[N&X3&!$(NC~[!#1&!#6]):3][C&X3:1]=[C&X3:2]",  # enamine
        "[O&X2&H1:3][C&X3:2]=[C:1]",  # enol
        "[C:1]1[C:1][O:2]1",  # epoxide
        "[#6:5][O&X2:2][C&X3:1](=[O:3])[#6:4]",  # ester
        "[O&D2&!$(OC~[!#1&!#6]):2]([#6:1])[#6:1]",  # ether
        "[F&-:1]",  # fluoride
        "[F,Cl,Br,I:1]",  # halogen
        "[C:4][O&X2:3][C&H1:1]([#6:2])[O&H1&X2:5]",  # hemiacetal
        "[!#1&!#6:1]",  # hetero
        "[#6:1][N&X3:2][N&X3:3]",  # hydrazine
        "[N&X3:3][N&X2:2]=[C:1]",  # hydrazone
        "[#6:1][O&H1&!$(O[#6]~[!#1&!#6]):2]",  # hydroxyl
        "[N&X2:2][O&H1:3]",  # hydroxylamine
        "[C&X3:1](=[O&X1:3])[N&X3:2][C&X3:1]=[O&X1:3]",  # imide
        "[C&X3:1]=[N&X2&!$(N~C~[!#1&!#6]):3][#6,#1:2]",  # imine
        "[N&X3&+:2]=[C&X3:1]",  # iminium
        "[I&-:1]",  # iodide
        "[N&X2:2]=[C:1]=[O&X1:3]",  # isocyanate
        "[N&+:2]#[C&-:1]",  # isocyanide
        "[N&+:2]#[C&-:1]",  # isonitrile
        "[N&X2:2]=[C:1]=[S&X1:3]",  # isothiocyanate
        "[O&X2:2][C&H0&X4&!$(CO~[!#6]):1][O&X2:3]",  # ketal
        "[C:1]=[C:2]=[O&X1:3]",  # ketene
        "[#6:2][C&X3:1](=[O:3])[#6:2]",  # ketone
        "[C:1](=[N:2][O&H1:3])([#6:4])[#6:4]",  # ketoxime
        # metal
        "[!#1&!#2&!#5&!#6&!#7&!#8&!#9&!#10&!#14&!#15&!#16&!#17&!#18&!#32&!#33&!#34&!#35&!#36&!#51&!#52&!#53&!#54&!#85&!#86:1]",
        "[B,#14,#32,#33,#51,#52,#85:1]",  # metalloid
        "[#7:1][O&X1&!$(O[#7]~[!#1&!#6]):2]",  # N oxide
        "[N&X1:2]#[C&X2:1]",  # nitrile
        "[#6:1][$([N&X3:2](=[O:3])=[O:3]),$([N&X3&+:1](=[O:3])[O&-:3])]",  # nitro
        "[N&X2:1]=[O&X1:2]",  # nitroso
        "[#1,#2,#6,#7,#8,F,#10,#15,#16,Cl,#18,#34,Br,#36,I,#54,#86:1]",  # nonmetal
        "[C:1]=[N:2][O&H1:3]",  # oxime
        "[O&X2,O&X1&-][O&X2,O&X1&-]",  # peroxide
        "[c:1][O&X2&H1:2]",  # phenol
        "[c:1][O&X1&-:2]",  # phenoxide
        "[c:1][O&X2&H0&!$(OC~[!#1&!#6]):2]",  # phenoxy
        "[c&H1]1[c&H1][c&H1][c&H1][c&H1]c1",  # phenyl
        "[PX3;$([H3]),$([H2][#6]),$([H1]([#6])[#6]),$([H0]([#6])([#6])[#6])]",  # phosphine
        "[N&X3&H2:2][C&X3:1]=[O&X1:3]",  # primary amide
        "[#6:1][#7&A&H2&X3&!$(NC~[!#1&!#6&!F&!Cl&!Br&!I])&!$(N~[!#1&!#6]):2]",  # primary amine
        "[H1,H2,H3,H4&+:1]",  # protonated
        "[#6&X4:1][N&X4&H0&!$(NC~[!#1&!#6]):2]",  # quaternary amino
        "[R:1]",  # ring
        "[N&X3&H1:2][C&X3:1]=[O&X1:3]",  # secondary amide
        "[#6:1][N&X3&H1&!$(NC~[!#1&!#6&!F&!Cl&!Br&!I])&!$(N~[!#1&!#6]):2]",  # secondary amine
        "[#6:1][S&X2&H0&!$(SC~[!#1&!#6]):2][#6:1]",  # sulfide
        # sulfonamide
        "[#6:1][$([S&X4:2]([N&X3:3])(=[O&X1:4])=[O&X1:4]),$([S&X4&+2:2]([N&X3:3])([O&X1&-:4])[O&X1&-:4])]",
        # sulfone
        "[$([S&X4:2](=[O&X1:3])(=[O&X1:3])([#6:1])[#6:1]),$([S&X4&+2:2]([O&X1&-:3])([O&X1&-:3])([#6:1])[#6:1])]",
        # sulfonic acid
        "[#6:1][$([S&X4:2](=[O&X1:4])(=[O&X1:4])[O&X2&H1,O&X1&H0&-:3]),$([S&X4&+2:2]([O&X1&-:4])([O&X1&-:4])[O&X2&H1,O&X1&H0&-:3])]",
        "[$([SX3](=[OX1])([#6])[#6]),$([SX3+]([OX1-])([#6])[#6])]",  # sulfoxide
        "[N&X3&H0:2][C&X3:1]=[O&X1:3]",  # tertiary amide
        "[#6:1][N&X3&H0&!$(NC~[!#1&!#6&!F&!Cl&!Br&!I]):2]",  # tertiary amine
        "[#6:1][N&X3&H0&!$(NC~[!#1&!#6&!F&!Cl&!Br&!I]):2]",  # tertiary amino
        "[N&X3:2][C&X3:1]=[S&X1:3]",  # thioamide
        "[C&X3:1](=[S:2])[S&X2&H1:3]",  # thiocarboxide
        "[C&X3:1](=[S:2])[S&-:3]",  # thiocarboxylate
        "[#6:5][S&X2:2][C&X3:1](=[O:3])[#6:4]",  # thioester
        "[#6:1][S&H1&!$(S[#6]~[!#1&!#6]):2]",  # thiol
        "[N&X3:2][C&X3:1](=[S&X1:3])[N&X3:2]",  # thiourea
        "[N&X3:2][C&X3:1](=[O&X1:3])[N&X3:2]",  # urea
        "[C&X3&H2:1]=[C&X3&H1:1]",  # vinyl
    ]
    # flake8: noqa
    fragment_smarts = [MolFromSmarts(smarts) for smarts in fragment_smarts]
    return fragment_smarts



import os
from collections.abc import Iterator
from functools import partial
from typing import Optional, Union

import numpy as np
import pandas as pd
from sklearn.utils._param_validation import StrOptions, validate_params

from skfp.datasets.lrgb.peptides_func import load_peptides_func
from skfp.datasets.lrgb.peptides_struct import load_peptides_struct
from skfp.datasets.utils import fetch_splits


@validate_params(
    {
        "data_dir": [None, str, os.PathLike],
        "mol_type": [StrOptions({"SMILES", "aminoseq"})],
        "standardize_labels": ["boolean"],
        "as_frame": ["boolean"],
        "verbose": ["boolean"],
    },
    prefer_skip_nested_validation=True,
)
def load_lrgb_mol_benchmark(
    data_dir: Optional[Union[str, os.PathLike]] = None,
    mol_type: str = "SMILES",
    standardize_labels: bool = True,
    as_frames: bool = False,
    verbose: bool = False,
) -> Union[
    Iterator[tuple[str, pd.DataFrame]], Iterator[tuple[str, list[str], np.ndarray]]
]:
    """
    Load and return the LRGB molecular datasets.

    There are two datasets: Peptides-func (binary multitask classification) and
    Peptides-struct (multitask regression). Stratified random split is recommended for
    both, following LRGB [1]_. See paper for details on stratification. AUPRC metric
    is recommended for Peptides-func, and MAE for Peptides-struct.

    Dataset names are also returned (case-sensitive): "Peptides-func" and "Peptides-struct".

    Parameters
    ----------
    data_dir : {None, str, path-like}, default=None
        Path to the root data directory. If ``None``, currently set scikit-learn directory
        is used, by default `$HOME/scikit_learn_data`.

    mol_type : {"SMILES", "aminoseq"}, default="SMILES"
        Which molecule representation to return, either SMILES strings or aminoacid
        sequences.

    standardize_labels : bool, default=True
        Whether to standardize labels to mean 0 and standard deviation 1 for
        Peptides-struct, following the recommendation from the original paper [1]_.
        Otherwise, the raw property values are returned.

    as_frames : bool, default=False
        If True, returns the raw DataFrame for each dataset. Otherwise, returns SMILES
        as a list of strings, and labels as a NumPy array for each dataset.

    verbose : bool, default=False
        If True, progress bar will be shown for downloading or loading files.

    Returns
    -------
    data : generator of pd.DataFrame or tuples (list[str], np.ndarray)
        Loads and returns datasets with a generator. Returned types depend on the
        ``as_frame`` and ``mol_type`` parameters, either:
        - Pandas DataFrame with columns: "SMILES"/"aminoseq", "label"
        - tuple of: list of strings (SMILES / aminoacid sequences), NumPy array (labels)

    References
    ----------
    .. [1] `Dwivedi, Vijay Prakash, et al.
        "Long Range Graph Benchmark"
        Advances in Neural Information Processing Systems 35 (2022): 22326-22340
        <https://proceedings.neurips.cc/paper_files/paper/2022/hash/8c3c666820ea055a77726d66fc7d447f-Abstract-Datasets_and_Benchmarks.html>`_
    """
    datasets = [
        (
            "Peptides-func",
            load_peptides_func,
        ),
        (
            "Peptides-struct",
            partial(load_peptides_struct, standardize_labels=standardize_labels),
        ),
    ]
    load_args = {
        "data_dir": data_dir,
        "mol_type": mol_type,
        "as_frame": as_frames,
        "verbose": verbose,
    }

    if as_frames:
        # generator of tuples (dataset_name, DataFrame)
        datasets_gen = (
            (dataset_name, load_function(**load_args))
            for dataset_name, load_function in datasets
        )
    else:
        # generator of tuples (dataset_name, SMILES/aminoseq, y)
        datasets_gen = (
            (dataset_name, *load_function(**load_args))
            for dataset_name, load_function in datasets
        )

    return datasets_gen


@validate_params(
    {
        "dataset_name": [StrOptions({"Peptides-func", "Peptides-struct"})],
        "data_dir": [None, str, os.PathLike],
        "as_frame": ["boolean"],
        "verbose": ["boolean"],
    },
    prefer_skip_nested_validation=True,
)
def load_lrgb_mol_splits(
    dataset_name: str,
    data_dir: Optional[Union[str, os.PathLike]] = None,
    as_dict: bool = False,
    verbose: bool = False,
) -> Union[tuple[list[int], list[int], list[int]], dict[str, list[int]]]:
    """
    Load and return the official LRGB splits for molecular datasets.

    Long Range Graph Benchmark (LRGB) [1]_ uses precomputed stratified random split for
    both Peptides-func and Peptides-struct datasets.

    Dataset names here are the same as returned by `load_moleculenet_benchmark` function,
    and are case-sensitive.

    Parameters
    ----------
    dataset_name : {"Peptides-func", "Peptides-struct"}
        Name of the dataset to loads splits for.

    data_dir : {None, str, path-like}, default=None
        Path to the root data directory. If ``None``, currently set scikit-learn directory
        is used, by default `$HOME/scikit_learn_data`.

    as_dict : bool, default=False
        If True, returns the splits as dictionary with keys "train", "valid" and "test",
        and index lists as values. Otherwise, returns three lists with splits indexes.

    verbose : bool, default=False
        If True, progress bar will be shown for downloading or loading files.

    Returns
    -------
    data : tuple(list[int], list[int], list[int]) or dict
        Depending on the `as_dict` argument, one of:
        - three lists of integer indexes
        - dictionary with "train", "valid" and "test" keys, and values as lists with
        splits indexes

    References
    ----------
    .. [1] `Dwivedi, Vijay Prakash, et al.
        "Long Range Graph Benchmark"
        Advances in Neural Information Processing Systems 35 (2022): 22326-22340
        <https://proceedings.neurips.cc/paper_files/paper/2022/hash/8c3c666820ea055a77726d66fc7d447f-Abstract-Datasets_and_Benchmarks.html>`_
    """
    splits = fetch_splits(
        data_dir,
        dataset_name=f"LRGB_{dataset_name}",
        filename=f"lrgb_splits_{dataset_name.lower().replace('-', '_')}.json",
        verbose=verbose,
    )
    if as_dict:
        return splits
    else:
        return splits["train"], splits["valid"], splits["test"]



from abc import ABC, abstractmethod
from collections.abc import Sequence
from copy import deepcopy
from numbers import Integral
from typing import Optional, Union

import numpy as np
from joblib import effective_n_jobs
from rdkit.Chem import Mol
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils._param_validation import InvalidParameterError
from tqdm import tqdm

from skfp.utils import ensure_mols, run_in_parallel


class BaseFilter(ABC, BaseEstimator, TransformerMixin):
    """Base class for molecular filters."""

    # parameters common for all filters
    _parameter_constraints: dict = {
        "allow_one_violation": ["boolean"],
        "return_indicators": ["boolean"],
        "n_jobs": [Integral, None],
        "batch_size": [Integral, None],
        "verbose": ["verbose"],
    }

    def __init__(
        self,
        allow_one_violation: bool = True,
        return_indicators: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        self.allow_one_violation = allow_one_violation
        self.return_indicators = return_indicators
        self.n_jobs = n_jobs
        self.batch_size = batch_size
        self.verbose = verbose

    def __sklearn_is_fitted__(self) -> bool:
        return True

    def fit(
        self, X: Sequence[Union[str, Mol]], y: Optional[np.ndarray] = None, **fit_params
    ):
        """Unused, kept for Scikit-learn compatibility.

        Parameters
        ----------
        X : any
            Unused, kept for Scikit-learn compatibility.

        y : any
            Unused, kept for Scikit-learn compatibility.

        **fit_params : dict
            Unused, kept for Scikit-learn compatibility.

        Returns
        --------
        self
        """
        self._validate_params()
        return self

    def fit_transform(
        self, X: Sequence[Union[str, Mol]], y: Optional[np.ndarray] = None, **fit_params
    ):
        """
        The same as ``.transform()`` method, kept for Scikit-learn compatibility.

        Parameters
        ----------
        X : any
            See ``.transform()`` method.

        y : any
            See ``.transform()`` method.

        **fit_params : dict
            Unused, kept for Scikit-learn compatibility.

        Returns
        -------
        X_new : any
            See ``.transform()`` method.
        """
        return self.transform(X)

    def transform(
        self, X: Sequence[Union[str, Mol]], copy: bool = False
    ) -> Union[list[Union[str, Mol]], np.ndarray]:
        """
        Apply a filter to input molecules. Output depends on ``return_indicators``
        attribute.

        Parameters
        ----------
        X : {sequence, array-like} of shape (n_samples,)
            Sequence containing RDKit Mol objects.

        copy : bool, default=False
            Copy the input X or not.

        Returns
        -------
        X : list of shape (n_samples_conf_gen,) or array of shape (n_samples,)
            List with filtered molecules, or indicator vector which molecules
            fulfill the filter rules.
        """
        filter_ind = self._get_filter_indicators(X, copy)
        if self.return_indicators:
            return filter_ind
        else:
            return [mol for idx, mol in enumerate(X) if filter_ind[idx]]

    def transform_x_y(
        self, X: Sequence[Union[str, Mol]], y: np.ndarray, copy: bool = False
    ) -> Union[tuple[list[Union[str, Mol]], np.ndarray], tuple[np.ndarray, np.ndarray]]:
        """
        Apply a filter to input molecules. Output depends on ``return_indicators``
        attribute.

        Parameters
        ----------
        X : {sequence, array-like} of shape (n_samples,)
            Sequence containing RDKit Mol objects.

        y : array-like of shape (n_samples,)
            Array with labels for molecules.

        copy : bool, default=False
            Copy the input X or not.

        Returns
        -------
        X : list of shape (n_samples_conf_gen,) or array of shape (n_samples,)
            List with filtered molecules, or indicator vector which molecules
            fulfill the filter rules.

        y : np.ndarray of shape (n_samples_conf_gen,)
            Array with labels for molecules.
        """
        filter_ind = self._get_filter_indicators(X, copy)
        mols = [mol for idx, mol in enumerate(X) if filter_ind[idx]]
        y = y[filter_ind]
        if self.return_indicators:
            return filter_ind, y
        else:
            return mols, y

    def _get_filter_indicators(
        self, mols: Sequence[Union[str, Mol]], copy: bool
    ) -> np.ndarray:
        self._validate_params()
        mols = deepcopy(mols) if copy else mols
        mols = ensure_mols(mols)

        n_jobs = effective_n_jobs(self.n_jobs)
        if n_jobs == 1:
            if self.verbose:
                filter_indicators = [
                    self._filter_mols_batch([mol]) for mol in tqdm(mols)
                ]
            else:
                filter_indicators = self._filter_mols_batch(mols)
        else:
            filter_indicators = run_in_parallel(
                self._filter_mols_batch,
                data=mols,
                n_jobs=n_jobs,
                batch_size=self.batch_size,
                flatten_results=True,
                verbose=self.verbose,
            )

        return filter_indicators

    def _filter_mols_batch(self, mols: list[Mol]) -> np.ndarray:
        filter_indicators = [self._apply_mol_filter(mol) for mol in mols]
        return np.array(filter_indicators, dtype=bool)

    @abstractmethod
    def _apply_mol_filter(self, mol: Mol) -> bool:
        pass

    def _validate_params(self) -> None:
        # override Scikit-learn validation to make stacktrace nicer
        try:
            super()._validate_params()
        except InvalidParameterError as e:
            raise InvalidParameterError(str(e)) from None



from typing import Union

from rdkit.Chem import GetFormalCharge, Mol
from rdkit.Chem.Crippen import MolLogP
from rdkit.Chem.Descriptors import MolWt
from rdkit.Chem.rdMolDescriptors import (
    CalcNumHBA,
    CalcNumHBD,
    CalcNumHeteroatoms,
    CalcNumRotatableBonds,
    CalcTPSA,
)

from skfp.bases.base_filter import BaseFilter

from .utils import (
    get_max_num_fused_aromatic_rings,
    get_max_ring_size,
    get_num_aromatic_rings,
    get_num_carbon_atoms,
    get_num_charged_atoms,
    get_num_heavy_metals,
    get_num_rigid_bonds,
)


class ValenceDiscoveryFilter(BaseFilter):
    """
    Valence Discovery © generative design filter.

    Set of rules proprietary to Valence Discovery ©, curated to prioritize molecules
    for generative models. Based on definitions from MedChem [1]_.

    Molecule must fulfill conditions:

    - molecular weight in range ``[200, 600]``
    - logP in range ``[-3, 6]``
    - HBA <= 12
    - HBD <= 7
    - TPSA in range ``[40, 180]``
    - number of rotatable bonds <= 15
    - number of rigid bonds <= 30
    - number of aromatic rings <= 5
    - number of aromatic rings fused together <= 2
    - max ring size <= 18
    - number of heavy atoms < 70
    - number of heavy metals < 1
    - number of carbons in range ``[3, 40]``
    - number of heteroatoms in range ``[1, 15]``
    - total formal charge in range ``[-2, 2]``
    - number of charged atoms <= 2

    Heavy atoms are defined as metals other than ["Li", "Be", "K", "Na", "Ca", "Mg"].

    Parameters
    ----------
    allow_one_violation : bool, default=False
        Whether to allow violating one of the rules for a molecule. This makes the
        filter less restrictive.

    return_indicators : bool, default=False
        Whether to return a binary vector with indicators which molecules pass the
        filter, instead of list of molecules.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform_x_y` and
        :meth:`transform` are parallelized over the input molecules. ``None`` means 1
        unless in a :obj:`joblib.parallel_backend` context. ``-1`` means using all
        processors. See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when filtering molecules.

    References
    -----------
    .. [1] `MedChem documentation - rule_of_generative_design
        <https://medchem-docs.datamol.io/stable/api/medchem.rules.html#medchem.rules.basic_rules.rule_of_generative_design>`_

    Examples
    ----------
    >>> from skfp.filters import ValenceDiscoveryFilter
    >>> smiles = ["C", "ClC1=CC2=C(N=C(NC)C[N+]([O-])=C2C3=CC=CC=C3)C=C1"]
    >>> filt = ValenceDiscoveryFilter()
    >>> filt
    ValenceDiscoveryFilter()

    >>> filtered_mols = filt.transform(smiles)
    >>> filtered_mols
    ['ClC1=CC2=C(N=C(NC)C[N+]([O-])=C2C3=CC=CC=C3)C=C1']
    """

    def __init__(
        self,
        allow_one_violation: bool = False,
        return_indicators: bool = False,
        n_jobs: Union[int, None] = None,
        batch_size: Union[int, None] = None,
        verbose: int = 0,
    ):
        super().__init__(
            allow_one_violation, return_indicators, n_jobs, batch_size, verbose
        )

    def _apply_mol_filter(self, mol: Mol) -> bool:
        rules = [
            200 <= MolWt(mol) <= 600,
            -3 <= MolLogP(mol) <= 6,
            CalcNumHBA(mol) <= 12,
            CalcNumHBD(mol) <= 7,
            40 <= CalcTPSA(mol) <= 180,
            CalcNumRotatableBonds(mol) <= 15,
            get_num_rigid_bonds(mol) <= 30,
            get_num_aromatic_rings(mol) <= 5,
            get_max_num_fused_aromatic_rings(mol) <= 2,
            get_max_ring_size(mol) <= 18,
            mol.GetNumHeavyAtoms() < 70,
            get_num_heavy_metals(mol) < 1,
            3 <= get_num_carbon_atoms(mol) <= 40,
            1 <= CalcNumHeteroatoms(mol) <= 15,
            -2 <= GetFormalCharge(mol) <= 2,
            get_num_charged_atoms(mol) <= 2,
        ]
        passed_rules = sum(rules)

        if self.allow_one_violation:
            return passed_rules >= len(rules) - 1
        else:
            return passed_rules == len(rules)



from collections.abc import Sequence
from contextlib import nullcontext
from typing import Optional

import numpy as np
from rdkit.Chem import Mol, MolFromInchi, MolToInchi

from skfp.bases import BasePreprocessor
from skfp.utils import no_rdkit_logs
from skfp.utils.functions import get_data_from_indices
from skfp.utils.validators import check_mols, check_strings


class MolFromInchiTransformer(BasePreprocessor):
    """
    Creates RDKit ``Mol`` objects from InChI strings.

    For details see RDKit documentation [1]_.

    Parameters
    ----------
    sanitize : bool, default=True
        Whether to perform sanitization [1]_, i.e. basic validity checks, on created
        molecules.

    remove_hydrogens : bool, default=True
        Remove explicit hydrogens from the molecule where possible, using RDKit
        implicit hydrogens instead.

    valid_only: bool, default=False
        Whether to return only molecules that were successfully loaded. By default,
        returns ``None`` for molecules that got errors.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized
        over the input molecules. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    suppress_warnings: bool, default=False
        Whether to suppress warnings and errors on loading molecules.

    verbose : int, default=0
        Controls the verbosity when processing molecules.

    References
    ----------
    .. [1] `RDKit InChI documentation
        <https://www.rdkit.org/docs/source/rdkit.Chem.inchi.html>`_

    Examples
    --------
    >>> from skfp.preprocessing import MolFromInchiTransformer
    >>> inchi_list = ["1S/H2O/h1H2", "1S/C8H10N4O2/c1-10-4-9-6-5(10)7(13)12(3)8(14)11(6)2/h4H,1-3H3"]
    >>> mol_from_inchi = MolFromInchiTransformer()
    >>> mol_from_inchi
    MolFromInchiTransformer()

    >>> mol_from_inchi.transform(inchi_list)  # doctest: +SKIP
        [<rdkit.Chem.rdchem.Mol>,
         <rdkit.Chem.rdchem.Mol>]
    """

    _parameter_constraints: dict = {
        **BasePreprocessor._parameter_constraints,
        "sanitize": ["boolean"],
        "remove_hydrogens": ["boolean"],
        "valid_only": ["boolean"],
    }

    def __init__(
        self,
        sanitize: bool = True,
        remove_hydrogens: bool = True,
        valid_only: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        suppress_warnings: bool = False,
        verbose: int = 0,
    ):
        super().__init__(
            n_jobs=n_jobs,
            batch_size=batch_size,
            suppress_warnings=suppress_warnings,
            verbose=verbose,
        )
        self.sanitize = sanitize
        self.remove_hydrogens = remove_hydrogens
        self.valid_only = valid_only

    def transform(self, X, copy: bool = False) -> list[Mol]:
        """
        Create RDKit ``Mol`` objects from InChI strings. If ``valid_only`` is set to
        True, returns only a subset of molecules which could be successfullyloaded.

        Parameters
        ----------
        X : {sequence, array-like} of shape (n_samples,)
            Sequence containing InChI strings.

        copy : bool, default=False
            Unused, kept for Scikit-learn compatibility.

        Returns
        -------
        X : list of shape (n_samples_conf_gen,)
            List with RDKit ``Mol`` objects.
        """
        X = super().transform(X, copy)
        if self.valid_only:
            X = [mol for mol in X if mol is not None]
        return X

    def transform_x_y(self, X, y, copy: bool = False) -> tuple[list[Mol], np.ndarray]:
        """
        Create RDKit ``Mol`` objects from InChI strings. If ``valid_only`` is set to
        True, returns only a subset of molecules and labels which could be successfully
        loaded.

        Parameters
        ----------
        X : {sequence, array-like} of shape (n_samples,)
            Sequence containing InChI strings

        y : np.ndarray of shape (n_samples,)
            Array with labels for molecules.

        copy : bool, default=False
            Unused, kept for Scikit-learn compatibility.

        Returns
        -------
        X : list of shape (n_samples,)
            List with RDKit ``Mol`` objects.

        y : np.ndarray of shape (n_samples,)
            Array with labels for molecules.
        """
        X = super().transform(X, copy)
        if self.valid_only:
            idxs_to_keep = [idx for idx, mol in X if mol is not None]
            X = get_data_from_indices(X, idxs_to_keep)
            y = y[idxs_to_keep]

        return X, y

    def _transform_batch(self, X: Sequence[str]) -> list[Mol]:
        with no_rdkit_logs() if self.suppress_warnings else nullcontext():
            check_strings(X)
            return [
                MolFromInchi(
                    inchi, sanitize=self.sanitize, removeHs=self.remove_hydrogens
                )
                for inchi in X
            ]


class MolToInchiTransformer(BasePreprocessor):
    """
    Creates InChI strings from RDKit ``Mol`` objects.

    For details see RDKit documentation [1]_.

    Parameters
    ----------
    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized
        over the input molecules. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when processing molecules.

    References
    ----------
    .. [1] `RDKit InChI documentation
        <https://www.rdkit.org/docs/source/rdkit.Chem.inchi.html>`_

    Examples
    --------
    >>> from skfp.preprocessing import MolFromInchiTransformer, MolToInchiTransformer
    >>> inchi_list = ["InChI=1S/H2O/h1H2", "InChI=1S/C8H10N4O2/c1-10-4-9-6-5(10)7(13)12(3)8(14)11(6)2/h4H,1-3H3"]
    >>> mol_from_inchi = MolFromInchiTransformer()
    >>> mol_to_inchi = MolToInchiTransformer()
    >>> mol_to_inchi
    MolToInchiTransformer()

    >>> mols = mol_from_inchi.transform(inchi_list)
    >>> mol_to_inchi.transform(mols)
    ['InChI=1S/H2O/h1H2', 'InChI=1S/C8H10N4O2/c1-10-4-9-6-5(10)7(13)12(3)8(14)11(6)2/h4H,1-3H3']
    """

    def __init__(
        self,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )

    def _transform_batch(self, X: Sequence[Mol]) -> list[str]:
        check_mols(X)
        return [MolToInchi(mol) for mol in X]



"""Classes for distance and similarity calculations."""

from .dice import (
    dice_binary_distance,
    dice_binary_similarity,
    dice_count_distance,
    dice_count_similarity,
)
from .tanimoto import (
    tanimoto_binary_distance,
    tanimoto_binary_similarity,
    tanimoto_count_distance,
    tanimoto_count_similarity,
)



import itertools
from collections.abc import Sequence
from typing import Callable, Optional

from joblib import effective_n_jobs
from sklearn.utils.parallel import Parallel, delayed
from tqdm import tqdm


class ProgressParallel(Parallel):
    """
    A more verbose version of ``joblib.Parallel``, which outputs progress bar.

    Parameters
    ----------
    total : int, default=None
        Total number of inputs to process in parallel.
    """

    def __init__(self, total: Optional[int] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.total = total

    def __call__(self, *args, **kwargs):
        with tqdm(total=self.total) as self._pbar:
            return Parallel.__call__(self, *args, **kwargs)

    def print_progress(self) -> None:
        self._pbar.n = self.n_completed_tasks
        self._pbar.refresh()


def run_in_parallel(
    func: Callable,
    data: Sequence,
    n_jobs: Optional[int] = None,
    batch_size: Optional[int] = None,
    flatten_results: bool = False,
    verbose: int = 0,
) -> list:
    """Run a function in parallel on provided data in batches, using joblib.

    Results are returned in the same order as input data. ``func`` function must take
    batch of data, e.g. list of integers, not a single integer.

    If ``func`` returns lists, the result will be a list of lists. To get a flat list
    of results, use ``flatten_results=True``.

    Note that progress bar for ``verbose`` option tracks processing of data batches,
    not individual data points.

    Parameters
    ----------
    func : Callable
        The function to run in parallel. It must take only a single argument,
        a batch of data.

    data : {sequence, array-like} of shape (n_samples,)
        Sequence containing data to process.

    n_jobs : int, default=None
        The number of jobs to run in parallel. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    flatten_results : bool, default=False
        Whether to flatten the results, e.g. to change list of lists of integers
        into a list of integers.

    verbose : int, default=0
        Controls the verbosity. If higher than zero, progress bar will be shown,
        tracking the processing of batches.

    Returns
    -------
    X : list of length (n_samples,)
        The processed data. If processing function returns functions, this will be
        a list of lists.

    Examples
    --------
    >>> from skfp.utils import run_in_parallel
    >>> func = lambda X: [x + 1 for x in X]
    >>> data = list(range(10))
    >>> run_in_parallel(func, data, n_jobs=-1, batch_size=1)
    [[1], [2], [3], [4], [5], [6], [7], [8], [9], [10]]
    >>> run_in_parallel(func, data, n_jobs=-1, batch_size=1, flatten_results=True)
    [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    """
    n_jobs = effective_n_jobs(n_jobs)

    if batch_size is None:
        batch_size = max(len(data) // n_jobs, 1)
    elif batch_size <= 0:
        raise ValueError(f"batch_size must be positive, got {batch_size}")

    data_batch_gen = (data[i : i + batch_size] for i in range(0, len(data), batch_size))
    num_batches = len(data) // batch_size

    if verbose > 0:
        parallel = ProgressParallel(n_jobs=n_jobs, total=num_batches)
    else:
        parallel = Parallel(n_jobs=n_jobs)

    results = parallel(delayed(func)(data_batch) for data_batch in data_batch_gen)

    if flatten_results:
        results = list(itertools.chain.from_iterable(results))

    return results



from .butina_split import butina_train_test_split, butina_train_valid_test_split
from .maxmin_split import maxmin_train_test_split, maxmin_train_valid_test_split
from .randomized_scaffold_split import (
    randomized_scaffold_train_test_split,
    randomized_scaffold_train_valid_test_split,
)
from .scaffold_split import scaffold_train_test_split, scaffold_train_valid_test_split



from collections.abc import Sequence
from typing import Optional, Union

import numpy as np
from rdkit.Chem import Mol
from scipy.sparse import csr_array

from skfp.bases import BaseFingerprintTransformer
from skfp.utils import ensure_mols, require_mols_with_conf_ids


class AutocorrFingerprint(BaseFingerprintTransformer):
    """
    Autocorrelation fingerprint.

    The implementation uses RDKit. This is a descriptor-based fingerprint, where
    bits measure strength of autocorrelation of molecular properties between atoms
    with different shortest path distances. For 3D variant, those distances are
    further weighted by Euclidean distance between atoms in a given conformation.

    It uses a hydrogen-depleted molecule, and for each heavy atom computes 6 features:
    atomic mass, van der Waals volume, electronegativity, polarizability, ion polarity,
    and IState [1]_ [2]_. They are then made relative to the carbon, e.g. molecular
    weight is: MW(atom_type) / MW(carbon).

    Four autocorrelation measures are used: Moreau-Broto, centered (average) Moreau-Broto,
    Moran and Geary [3]_ [4]_. They are calculated using topological distances (shortest
    paths), with distance between 1 and 8 (inclusive). This results in 192 features:
    6 atom features * 4 autocorrelations * 8 distances.

    3D variant has the following differences:

    - requires passing molecules with conformers and ``conf_id`` integer property set
    - weights topological distances by Euclidean distance between atoms
    - uses 2 additional features: constant 1 (which measures Euclidean distance
      autocorrelation due to weighting) and covalent radius (RCov)
    - uses shortest paths distances between 1 and 9 (inclusive)
    - uses only Moreau-Broto autocrrelation
    - results in 80 features: 8 atom features * 10 distances

    For more details on autocorrelation descriptors, see [5]_.

    Parameters
    ----------
    use_3D : bool, default=False
        Whether to use 3D Euclidean distance matrix. If False, only uses topological
        distances on molecular graph.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized
        over the input molecules. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when computing fingerprints.

    Attributes
    ----------
    n_features_out : int = 192 or 80
        Number of output features. Equal to 192 for 2D and 80 for 3D, which depends
        on the ``use_3D`` parameter.

    requires_conformers : bool
        Whether the fingerprint is 3D-based and requires molecules with conformers as
        inputs, with ``conf_id`` integer property set. This depends on the ``use_3D``
        parameter, and has the same value.

    References
    ----------
    .. [1] RDKit IState discussion
        https://github.com/rdkit/rdkit/discussions/6122

    .. [2] RDKit IState implementation
        https://github.com/rdkit/rdkit/blob/df2b6c8a8c775748c1dcac83af0f92af434bab81/Code/GraphMol/Descriptors/MolData3Ddescriptors.cpp#L127

    .. [3] `Rajarshi Guha
        "Autocorrelation Descriptors"
        <http://www.rguha.net/writing/notes/desc/node2.html>`_

    .. [4] `Roberto Todeschini and Viviana Consonni
        "Molecular Descriptors for Chemoinformatics"
        <https://onlinelibrary.wiley.com/doi/book/10.1002/9783527628766>`_

    .. [5] `Guillaume Godin
        "3D	descriptors in RDKit"
        UGM 2017
        <https://github.com/rdkit/UGM_2017/blob/master/Presentations/Godin_3D_Descriptors.pdf>`_

    Examples
    --------
    >>> from skfp.fingerprints import AutocorrFingerprint
    >>> smiles = ["CCO", "CCN"]
    >>> fp = AutocorrFingerprint()
    >>> fp
    AutocorrFingerprint()

    >>> fp.transform(smiles)  # doctest: +ELLIPSIS
    array([[ 1.204,  0.847,  0.   ,  ...,  0.   ,  0.   ,  0.   ],
           [ 1.153,  0.773,  0.   ,  ...,  0.   ,  0.   ,  0.   ]])
    """

    _parameter_constraints: dict = {
        **BaseFingerprintTransformer._parameter_constraints,
        "use_3D": ["boolean"],
    }

    def __init__(
        self,
        use_3D: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        n_features_out = 80 if use_3D else 192
        super().__init__(
            n_features_out=n_features_out,
            requires_conformers=use_3D,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )
        self.use_3D = use_3D

    def _calculate_fingerprint(
        self, X: Sequence[Union[str, Mol]]
    ) -> Union[np.ndarray, csr_array]:
        from rdkit.Chem.rdMolDescriptors import CalcAUTOCORR2D, CalcAUTOCORR3D

        if not self.use_3D:
            X = ensure_mols(X)
            X = [CalcAUTOCORR2D(mol) for mol in X]
        else:
            X = require_mols_with_conf_ids(X)
            X = [CalcAUTOCORR3D(mol, confId=mol.GetIntProp("conf_id")) for mol in X]

        return np.array(X)



from typing import Optional

from rdkit.Chem import Mol
from rdkit.Chem.Crippen import MolLogP
from rdkit.Chem.Descriptors import MolWt
from rdkit.Chem.rdMolDescriptors import CalcNumLipinskiHBA, CalcNumLipinskiHBD

from skfp.bases.base_filter import BaseFilter


class LipinskiFilter(BaseFilter):
    """
    Lipinski's Rule of 5 (RO5).

    Also known as Pfizer's Rule of 5. It evaluates the drug-likeness of a molecule
    as an orally active drug. Assumes that it should be small and lipophilic.
    Description of the rules can be found in the original publication [1]_.

    Molecule can violate at most one of the rules (conditions):
    - molecular weight <= 500 daltons
    - HBA <= 10
    - HBD <= 5
    - logP <= 5

    Hydrogen bond acceptors (HBA) and donors (HBD) use a simplified definition,
    taking into consideration only oxygen and nitrogen bonds with hydrogen (OH, NH).

    Parameters
    ----------
    allow_one_violation : bool, default=True
        Whether to allow violating one of the rules for a molecule. This makes the
        filter less restrictive, and is the part of the original definition of this
        filter.

    return_indicators : bool, default=False
        Whether to return a binary vector with indicators which molecules pass the
        filter, instead of list of molecules.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform_x_y` and
        :meth:`transform` are parallelized over the input molecules. ``None`` means 1
        unless in a :obj:`joblib.parallel_backend` context. ``-1`` means using all
        processors. See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when filtering molecules.

    References
    ----------
    .. [1] `Christopher A. Lipinski, Franco Lombardo, Beryl W. Dominy, Paul J. Feeney
        "Experimental and computational approaches to estimate solubility and permeability
        in drug discovery and development settings"
        Advanced Drug Delivery Reviews, Volume 23, Issues 1–3, 15 January 1997, Pages 3-25
        <https://www.sciencedirect.com/science/article/pii/S0169409X96004231>`_

    Examples
    --------
    >>> from skfp.filters import LipinskiFilter
    >>> smiles = ["[C-]#N", "CC=O", "O=C(O)c1ccccc1c2ccc(cc2)Cn3c4cc(cc(c4nc3CCC)C)c5nc6ccccc6n5C"]
    >>> filt = LipinskiFilter()
    >>> filt
    LipinskiFilter()

    >>> filtered_mols = filt.transform(smiles)
    >>> filtered_mols
    ['[C-]#N', 'CC=O']
    """

    def __init__(
        self,
        allow_one_violation: bool = True,
        return_indicators: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            allow_one_violation=allow_one_violation,
            return_indicators=return_indicators,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )

    def _apply_mol_filter(self, mol: Mol) -> bool:
        rules = [
            MolWt(mol) <= 500,  # molecular weight
            CalcNumLipinskiHBA(mol) <= 10,  # HBA
            CalcNumLipinskiHBD(mol) <= 5,  # HBD
            MolLogP(mol) <= 5,  # logP
        ]
        passed_rules = sum(rules)

        if self.allow_one_violation:
            return passed_rules >= len(rules) - 1
        else:
            return passed_rules == len(rules)



from typing import Optional

from rdkit.Chem import Mol
from rdkit.Chem.Crippen import MolLogP
from rdkit.Chem.rdMolDescriptors import CalcTPSA

from skfp.bases.base_filter import BaseFilter


class PfizerFilter(BaseFilter):
    """
    Pfizer 3/75 rule.

    Based on observation that compounds exhibiting low partition coefficient (clogP) and
    high topological polar surface area (TPSA) are roughly 2.5 times more likely to be
    free of toxicity issues in the tested conditions [1]_ [2]_.

    Molecule must fulfill conditions:

    - logP <= 3
    - TPSA >= 75

    Parameters
    ----------
    allow_one_violation : bool, default=False
        Whether to allow violating one of the rules for a molecule. This makes the
        filter less restrictive.

    return_indicators : bool, default=False
        Whether to return a binary vector with indicators which molecules pass the
        filter, instead of list of molecules.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform_x_y` and
        :meth:`transform` are parallelized over the input molecules. ``None`` means 1
        unless in a :obj:`joblib.parallel_backend` context. ``-1`` means using all
        processors. See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when filtering molecules.

    References
    -----------
    .. [1] `Hughes, J. D. et al.
        "Physiochemical drug properties associated with in vivo toxicological outcomes."
        Bioorganic & Medicinal Chemistry Letters, 18(17), 4872–4875.
        <https://doi.org/10.1016/j.bmcl.2008.07.071>`_

    .. [2] `Price, D. A., Blagg, J., Jones, L., Greene, N., & Wager, T.
        "Physicochemical drug properties associated with in vivo toxicological outcomes: a review."
        Expert Opinion on Drug Metabolism & Toxicology, 5(8), 921–931.
        <https://doi.org/10.1517/17425250903042318>`_

    Examples
    ----------
    >>> from skfp.filters import PfizerFilter
    >>> smiles = ["CS(=O)(=O)NCc1nnc(SCc2ccccc2C(F)(F)F)o1", "COC(=O)c1ccccc1NC(=O)CSc1nc(O)c(-c2ccccc2)c(=O)[nH]1",\
    "Cc1ccccc1-n1c(Cc2cccs2)n[nH]c1=S"]
    >>> filt = PfizerFilter()
    >>> filt
    PfizerFilter()
    >>> filtered_mols = filt.transform(smiles)
    >>> filtered_mols
    ['CS(=O)(=O)NCc1nnc(SCc2ccccc2C(F)(F)F)o1', 'COC(=O)c1ccccc1NC(=O)CSc1nc(O)c(-c2ccccc2)c(=O)[nH]1']
    """

    def __init__(
        self,
        allow_one_violation: bool = False,
        return_indicators: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            allow_one_violation=allow_one_violation,
            return_indicators=return_indicators,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )

    def _apply_mol_filter(self, mol: Mol) -> bool:
        rules = [
            MolLogP(mol) <= 3,
            CalcTPSA(mol) >= 75,
        ]
        passed_rules = sum(rules)

        if self.allow_one_violation:
            return passed_rules >= len(rules) - 1
        else:
            return passed_rules == len(rules)



from typing import Union

import numpy as np
from scipy.sparse import csr_array


def _check_nan(arr: Union[np.ndarray, csr_array]) -> None:
    if isinstance(arr, np.ndarray):
        if np.isnan(arr).any():
            raise ValueError("Input array contains NaN values")
    elif isinstance(arr, csr_array):
        if np.isnan(arr.data).any():
            raise ValueError("Input sparse matrix contains NaN values")
    else:
        raise TypeError(
            f"Expected numpy.ndarray or scipy.sparse.csr_array, got {type(arr)}"
        )



from collections.abc import Sequence
from copy import deepcopy
from typing import Optional, Union

import numpy as np
from rdkit.Chem import Mol
from scipy.sparse import csr_array
from sklearn.utils._param_validation import StrOptions

from skfp.bases import BaseFingerprintTransformer
from skfp.utils import require_mols_with_conf_ids


class USRFingerprint(BaseFingerprintTransformer):
    """
    USR (Ultrafast Shape Recognition) fingerprint.

    The implementation uses RDKit. This is a descriptor-based fingerprint, which
    characterizes the shape of the molecule by encoding the relative positions of its
    atoms [1]_ [2]_.

    Four points are considered: molecular centroid (ctd), the closest atom to centroid (cst),
    the farthest atom from centroid (fct), and atom the fartest from fct (ftf). Distances
    from all atoms to each of those four points are computed, and each of those distributions
    is summarized its first three moments: mean, variance, and skewness. Concretely, standard
    deviation and cubic root of skewness are used to keep the same unit. This results in
    12 descriptors.

    This is a 3D fingerprint, and requries molecules with ``conf_id`` integer property
    set. They can be generated with :class:`~skfp.preprocessing.ConformerGenerator`.
    Furthermore, only molecules with 3 or more atoms are allowed, to allow computation
    of all three moments.

    Parameters
    ----------
    errors : {"raise", "NaN", "ignore"}, default="raise"
        How to handle errors during fingerprint calculation. ``"raise"`` immediately
        raises any errors. ``"NaN"`` returns NaN values for molecules which resulted in
        errors. ``"ignore"`` suppresses errors and does not return anything for
        molecules with errors. This potentially results in less output vectors than
        input molecules, and should be used with caution.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized
        over the input molecules. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when computing fingerprints.

    Attributes
    ----------
    n_features_out : int = 12
        Number of output features, size of fingerprints.

    requires_conformers : bool = True
        Value is always True, as this fingerprint is 3D based. It always requires
        molecules with conformers as inputs, with ``conf_id`` integer property set.

    See Also
    --------
    :class:`USRCAT` : Related fingerprint, which additionally uses pharmacophoric
        atom types.

    :class:`ElectroShaoe` : Related fingerprint, which additionally uses atomic
        partial charges.

    References
    ----------
    .. [1] `Pedro J. Ballester and W. Graham Richards
        "Ultrafast shape recognition for similarity search in molecular databases"
        Proceedings of the Royal Society A: Mathematical, Physical and Engineering
        Sciences 463.2081 (2007): 1307-1321
        <https://royalsocietypublishing.org/doi/full/10.1098/rspa.2007.1823>`_

    .. [2] `Pedro J. Ballester
        "Ultrafast shape recognition: method and applications"
        Future Medicinal Chemistry 3.1 (2011): 65-78
        <https://www.future-science.com/doi/abs/10.4155/fmc.10.280>`_

    Examples
    --------
    >>> from skfp.fingerprints import USRFingerprint
    >>> from skfp.preprocessing import MolFromSmilesTransformer, ConformerGenerator
    >>> smiles = ["CC=O"]
    >>> fp = USRFingerprint()
    >>> fp
    USRFingerprint()

    >>> mol_from_smiles = MolFromSmilesTransformer()
    >>> mols = mol_from_smiles.transform(smiles)
    >>> conf_gen = ConformerGenerator()
    >>> mols = conf_gen.transform(mols)
    >>> fp.transform(mols)  # doctest: +SKIP
    array([[ 1.33723405,  0.39526642, ...        ,  0.94315243, -0.78539494]])
    """

    _parameter_constraints: dict = {
        **BaseFingerprintTransformer._parameter_constraints,
        "errors": [StrOptions({"raise", "NaN", "ignore"})],
    }

    def __init__(
        self,
        errors: str = "raise",
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            n_features_out=12,
            requires_conformers=True,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )
        self.errors = errors

    def transform(
        self, X: Sequence[Union[str, Mol]], copy: bool = False
    ) -> Union[np.ndarray, csr_array]:
        """
        Compute USR fingerprints. If ``errors`` is set to ``"ignore"``, then in
        case of errors less than n_samples values may be returned.

        Parameters
        ----------
        X : {sequence, array-like} of shape (n_samples,)
            Sequence containing RDKit Mol objects, with conformers generated and
            ``conf_id`` integer property set.

        copy : bool, default=False
            Copy the input X or not.

        Returns
        -------
        X : {ndarray, sparse matrix} of shape (n_samples, self.fp_size)
            Array with fingerprints.
        """
        y = np.empty(len(X))
        X, _ = self.transform_x_y(X, y, copy=copy)
        return X

    def transform_x_y(
        self, X: Sequence[Mol], y: np.ndarray, copy: bool = False
    ) -> tuple[Union[np.ndarray, csr_array], np.ndarray]:
        """
        Compute USR fingerprints. If ``errors`` is set to ``"ignore"``, then in
        case of errors less than n_samples values may be returned. The returned
        values for X and y are properly synchronized.

        Parameters
        ----------
        X : {sequence, array-like} of shape (n_samples,)
            Sequence containing RDKit Mol objects, with conformers generated and
            ``conf_id`` integer property set.

        y : np.ndarray of shape (n_samples,)
            Array with labels for molecules.

        copy : bool, default=False
            Copy the inputs X and y or not.

        Returns
        -------
        X : {ndarray, sparse matrix} of shape (n_samples, self.fp_size)
            Array with fingerprints.

        y : np.ndarray of shape (n_samples,)
            Array with labels for molecules.
        """
        if copy:
            X = deepcopy(X)
            y = deepcopy(y)

        X = super().transform(X)

        if self.errors == "ignore":
            # errors are marked as NaN rows
            idxs_to_keep = [idx for idx, x in enumerate(X) if not np.any(np.isnan(x))]
            X = X[idxs_to_keep]
            y = y[idxs_to_keep]

        return X, y

    def _calculate_fingerprint(self, X: Sequence[Mol]) -> Union[np.ndarray, csr_array]:
        from rdkit.Chem.rdMolDescriptors import GetUSR

        X = require_mols_with_conf_ids(X)

        if self.errors == "raise":
            fps = [GetUSR(mol, confId=mol.GetIntProp("conf_id")) for mol in X]
        else:  # self.errors in {"NaN", "ignore"}
            fps = []
            for mol in X:
                try:
                    fp = GetUSR(mol, confId=mol.GetIntProp("conf_id"))
                except ValueError:
                    fp = np.full(self.n_features_out, np.NaN)
                fps.append(fp)

        return np.array(fps)



from collections.abc import Sequence
from pathlib import Path
from typing import Optional, Union

import numpy as np
from rdkit.Chem import Mol
from scipy.sparse import csr_array

from skfp.bases import BaseSubstructureFingerprint
from skfp.utils import ensure_mols


class LaggnerFingerprint(BaseSubstructureFingerprint):
    """
    Substructure fingerprint with definitions by Christian Laggner.

    A substructure fingerprint based on SMARTS patterns for functional group
    classification, proposed by Christian Laggner [1]_. It is also known as
    SubstructureFingerprint in Chemistry Development Kit (CDK) [2]_.
    It tests for presence of 307 predefined substructures, designed for functional
    groups of organic compounds, for use in similarity searching.

    Parameters
    ----------
    count : bool, default=False
        Whether to return binary (bit) features, or their counts.

    sparse : bool, default=False
        Whether to return dense NumPy array, or sparse SciPy CSR array.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized
        over the input molecules. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when computing fingerprints.

    Attributes
    ----------
    n_features_out : int = 307
        Number of output features, size of fingerprints.

    requires_conformers : bool = False
        This fingerprint uses only 2D molecular graphs and does not require conformers.

    References
    ----------
    .. [1] `Christian Laggner
        "SMARTS Patterns for Functional Group Classification"
        OpenBabel
        <https://raw.githubusercontent.com/openbabel/openbabel/master/data/SMARTS_InteLigand.txt>`_

    .. [2] `egonw
        "SubstructureFingerprinter"
        Chemistry Development Kit (CDK) API reference
        <https://cdk.github.io/cdk/1.5/docs/api/org/openscience/cdk/fingerprint/SubstructureFingerprinter.html>`_

    Examples
    --------
    >>> from skfp.fingerprints import LaggnerFingerprint
    >>> smiles = ["O", "CC", "[C-]#N", "CC=O"]
    >>> fp = LaggnerFingerprint()
    >>> fp
    LaggnerFingerprint()

    >>> fp.transform(smiles)
    array([[0, 0, 0, ..., 0, 0, 0],
           [1, 0, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 0, 0, 0],
           [1, 0, 0, ..., 1, 0, 1]], dtype=uint8)
    """

    def __init__(
        self,
        count: bool = False,
        sparse: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        patterns = self._load_patterns()
        super().__init__(
            patterns=patterns,
            count=count,
            sparse=sparse,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )

    def transform(
        self, X: Sequence[Union[str, Mol]], copy: bool = False
    ) -> Union[np.ndarray, csr_array]:
        """
        Compute Laggner fingerprints.

        Parameters
        ----------
        X : {sequence, array-like} of shape (n_samples,)
            Sequence containing SMILES strings or RDKit Mol objects.

        copy : bool, default=False
            Copy the input X or not.

        Returns
        -------
        X : {ndarray, sparse matrix} of shape (n_samples, 307)
            Array with fingerprints.
        """
        return super().transform(X, copy)

    def _calculate_fingerprint(
        self, X: Sequence[Union[str, Mol]]
    ) -> Union[np.ndarray, csr_array]:
        from rdkit.Chem import GetMolFrags

        X = ensure_mols(X)
        fps = super()._calculate_fingerprint(X)

        # temporarily convert to dictionary-of-keys (DOK) format to set bits
        # for set  cation and salt features (setting bits on CSR array is slow)
        if self.sparse:
            fps = fps.todok()

        for idx, mol in enumerate(X):
            # salt = at least two components, has anion and cation
            # this can only be 0 or 1
            multi_component = len(GetMolFrags(mol)) > 1
            anion = fps[idx, 296]
            cation = fps[idx, 297]
            salt = multi_component & anion & cation
            fps[idx, 298] = salt

        return csr_array(fps) if self.sparse else fps

    def _load_patterns(self) -> list[str]:
        # since Laggner file is licensed under LGPL, we keep it separately
        patterns = []

        filepath = Path(__file__).parent / "data" / "SMARTS_InteLigand.txt"
        with open(filepath) as file:
            for line in file:
                if line.startswith("#") or line.isspace():
                    continue
                elif line.startswith("Urea:"):
                    smarts = line.removeprefix("Urea:")
                else:
                    name, smarts = line.split()
                patterns.append(smarts.strip())

        # RDKit does not support multi-component SMARTS (with a dot), so we can't match
        # the salts pattern "([-1,-2,-3,-4,-5,-6,-7]).([+1,+2,+3,+4,+5,+6,+7])"
        # (index 298); we compute it manually in .transform(), and here temporarily
        # replace it with an empty pattern
        patterns[298] = ""

        return patterns



from collections.abc import Sequence
from numbers import Real
from typing import Optional, Union

import numpy as np
from rdkit.Chem import Mol
from scipy.sparse import csr_array
from sklearn.utils._param_validation import Interval

from skfp.bases import BaseFingerprintTransformer
from skfp.utils import ensure_mols, ensure_smiles, no_rdkit_logs


class RDKit2DDescriptorsFingerprint(BaseFingerprintTransformer):
    """
    RDKit 2D descriptors fingerprint.

    The implementation uses descriptastorus [1]_ and RDKit. This fingerprint consists
    of 200 2D descriptors available in RDKit (almost all). List of all features is
    available in descriptastorus code and in the supplementary material of the original
    paper[2]_.

    Normalized variant uses cumulative distribution function (CDF) normalization, as
    proposed in [2]_. Distributions for normalization have been determined using a
    large collection of molecules from ChEMBL [3]_.

    Typical correct values should be small, but it often results in NaN or infinity
    for some descriptors. Value clipping with ``clip_val`` parameter, feature selection,
    and/or imputation should be used.

    Parameters
    ----------
    normalized : bool, default=False
        Whether to return CDF-normalized descriptor values.

    clip_val : float or None, default=2147483647
        Value to clip results at, both positive and negative ones.The default value is
        the maximal value of 32-bit integer, but should often be set lower, depending
        on the application. ``None`` means that no clipping is applied.

    sparse : bool, default=False
        Whether to return dense NumPy array, or sparse SciPy CSR array.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized
        over the input molecules. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when computing fingerprints.

    Attributes
    ----------
    n_features_out : int = 200
        Number of output features, size of fingerprints.

    requires_conformers : bool = False
        This fingerprint uses only 2D molecular graphs and does not require conformers.

    References
    ----------
    .. [1] Descriptastorus
        https://github.com/bp-kelley/descriptastorus

    .. [2] `Kevin Yang et al.
        "Analyzing Learned Molecular Representations for Property Prediction"
        Journal of Chemical Information and Modeling 59.8 (2019): 3370-3388
        <https://pubs.acs.org/doi/10.1021/acs.jcim.9b00237>`_

    .. [3] Descriptastorus normalized descriptors discussion
        https://github.com/bp-kelley/descriptastorus/issues/3

    Examples
    --------
    >>> from skfp.fingerprints import RDKit2DDescriptorsFingerprint
    >>> smiles = ["O", "CC", "[C-]#N", "CC=O"]
    >>> fp = RDKit2DDescriptorsFingerprint()
    >>> fp
    RDKit2DDescriptorsFingerprint()

    >>> fp.transform(smiles)  # doctest: +SKIP
    array([[ 1.00000000e+00,  0.00000000e+00,  ...  0.00000000e+00,  3.27747673e-01]
           [ 1.00000000e+00,  1.00000000e+00,  ...  0.00000000e+00,  3.72785568e-01]
           [ 1.00000000e+00,  3.00000000e+00,  ...  0.00000000e+00,  3.44374359e-01]
           [ 1.00000000e+00,  2.18749619e+00,  ...  0.00000000e+00,  3.55007619e-01]], dtype=float32)
    """

    _parameter_constraints: dict = {
        **BaseFingerprintTransformer._parameter_constraints,
        "normalized": ["boolean"],
        "clip_val": [None, Interval(Real, 0, None, closed="left")],
    }

    def __init__(
        self,
        normalized: bool = False,
        clip_val: float = 2147483647,  # max int32 value
        sparse: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            n_features_out=200,
            sparse=sparse,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )
        self.normalized = normalized
        self.clip_val = clip_val

    def _calculate_fingerprint(
        self, X: Sequence[Union[str, Mol]]
    ) -> Union[np.ndarray, csr_array]:
        from descriptastorus.descriptors.rdDescriptors import RDKit2D
        from descriptastorus.descriptors.rdNormalizedDescriptors import (
            RDKit2DNormalized,
        )

        mols = ensure_mols(X)
        smiles = ensure_smiles(X)

        # turn off RDKit logs, since descriptastorus does not use MorganGenerator
        with no_rdkit_logs():
            gen = RDKit2DNormalized() if self.normalized else RDKit2D()
            X = [np.array(gen.calculateMol(mol, smi)) for mol, smi in zip(mols, smiles)]

        # clip values to float32 range
        X = [np.clip(x, -self.clip_val, self.clip_val) for x in X]

        if self.sparse:
            return csr_array(X, dtype=np.float32)
        else:
            return np.array(X, dtype=np.float32)



from collections.abc import Sequence
from typing import Optional, Union

import numpy as np
from rdkit.Chem import Mol
from scipy.sparse import csr_array

from skfp.bases import BaseFingerprintTransformer
from skfp.utils import require_mols_with_conf_ids


class MORSEFingerprint(BaseFingerprintTransformer):
    r"""
    MoRSE (Molecule Representation of Structures based on Electron diffraction)
    fingerprint.

    The implementation uses RDKit. This is a descriptor-based fingerprint, where
    features encode basic 3D atomic coordinates (molecule structure) with a
    constant-size vector, based on principles similar to electron diffraction.

    MoRSE features are scattered electron intensity values for different values
    of scattering ``s``. It uses general molecular transform which, for all atoms i and j
    with distance :math:`r_{ij}` in the molecule with N atoms, is defined as:

    .. math::

        RDF(r) = \sum_{i}^{N-1} \sum_{j > i}^N w_i * w_j * \frac{\sin(s * r_{ij})}{s * r_{ij}}

    7 weighting variants are used, unweighted and 6 based on atomic features:
    unweighted, atomic mass, van der Waals volume, electronegativity, polarizability,
    ion polarity, and IState [1]_ [2]_. They are relative to the carbon, e.g. molecular
    weight is: MW(atom_type) / MW(carbon).

    Scattering values between 0 and 31 (inclusive) are used, in angstroms (Å).
    This results in 224 features. See [3]_ [4]_ [5]_ [6]_ [7]_ for details.

    Parameters
    ----------
    sparse : bool, default=False
        Whether to return dense NumPy array, or sparse SciPy CSR array.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized
        over the input molecules. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when computing fingerprints.

    Attributes
    ----------
    n_features_out : int = 224
        Number of output features, size of fingerprints.

    requires_conformers : bool = True
        Value is always True, as this fingerprint is 3D based. It always requires
        molecules with conformers as inputs, with ``conf_id`` integer property set.

    See Also
    --------
    :class:`RDFFingerprint` : Related fingerprint, which uses radial distribution
        function (RDF) instead of scattered electron intensity.

    References
    ----------
    .. [1] RDKit IState discussion
        https://github.com/rdkit/rdkit/discussions/6122

    .. [2] RDKit IState implementation
        https://github.com/rdkit/rdkit/blob/df2b6c8a8c775748c1dcac83af0f92af434bab81/Code/GraphMol/Descriptors/MolData3Ddescriptors.cpp#L127

    .. [3] `DRAGON documentation
        "MoRSE descriptors"
        <http://chemgps.bmc.uu.se/help/dragonx/MoRSEdescriptors.html>`_

    .. [4] `Roberto Todeschini and Viviana Consonni
        "Molecular Descriptors for Chemoinformatics"
        <https://onlinelibrary.wiley.com/doi/book/10.1002/9783527628766>`_

    .. [5] `Guillaume Godin
        "3D	descriptors in RDKit"
        UGM 2017
        <https://github.com/rdkit/UGM_2017/blob/master/Presentations/Godin_3D_Descriptors.pdf>`_

    .. [6] `Jan H. Schuur, Paul Selzer, and Johann Gasteiger
        "The Coding of the Three-Dimensional Structure of Molecules by Molecular
        Transforms and Its Application to Structure-Spectra Correlations and Studies
        of Biological Activity"
        J. Chem. Inf. Comput. Sci. 1996, 36, 2, 334–344
        <https://pubs.acs.org/doi/abs/10.1021/ci950164c>`_

    .. [7] `Devinyak, Oleg, Dmytro Havrylyuk, and Roman Lesyk
        "3D-MoRSE descriptors explained"
        Journal of Molecular Graphics and Modelling 54 (2014): 194-203
        <https://www.sciencedirect.com/science/article/pii/S109332631400165X>`_

    Examples
    --------
    >>> from skfp.fingerprints import MORSEFingerprint
    >>> from skfp.preprocessing import MolFromSmilesTransformer, ConformerGenerator
    >>> smiles = ["O", "CC", "[C-]#N", "CC=O"]
    >>> fp = MORSEFingerprint()
    >>> fp
    MORSEFingerprint()

    >>> mol_from_smiles = MolFromSmilesTransformer()
    >>> mols = mol_from_smiles.transform(smiles)
    >>> conf_gen = ConformerGenerator()
    >>> mols = conf_gen.transform(mols)
    >>> fp.transform(mols)  # doctest: +SKIP
    array([[ 3.0000e+00,  2.3090e+00,  9.0900e-01, ..., -1.3000e-02,
            -8.3000e-02, -5.3000e-02],
           [ 2.8000e+01,  1.2408e+01, -8.1900e-01, ...,  3.5400e-01,
             4.8200e-01, -1.2000e-01],
           [ 1.0000e+00,  7.9500e-01,  3.2700e-01, ...,  8.6700e-01,
             1.2400e-01, -7.1200e-01],
           [ 2.1000e+01,  9.2430e+00, -4.3800e-01, ..., -5.8700e-01,
             2.7400e-01, -3.9000e-02]])
    """

    def __init__(
        self,
        sparse: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            n_features_out=224,
            requires_conformers=True,
            sparse=sparse,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )

    def _calculate_fingerprint(self, X: Sequence[Mol]) -> Union[np.ndarray, csr_array]:
        from rdkit.Chem.rdMolDescriptors import CalcMORSE

        X = require_mols_with_conf_ids(X)
        X = [CalcMORSE(mol, confId=mol.GetIntProp("conf_id")) for mol in X]
        return csr_array(X) if self.sparse else np.array(X)



from typing import Optional

from rdkit.Chem import Mol
from rdkit.Chem.Crippen import MolLogP
from rdkit.Chem.Descriptors import MolWt
from rdkit.Chem.rdMolDescriptors import CalcNumHBA, CalcNumRings

from skfp.bases.base_filter import BaseFilter


class RuleOfFourFilter(BaseFilter):
    """
    Rule of four (Ro4).

    Rule designed to look for molecules used as PPI (protein-protein inhibitor).
    Described in [1]_.

    Molecule must fulfill conditions:

    - molecular weight >= 400 daltons
    - HBA >= 4
    - logP >=4
    - number of rings >= 4

    Parameters
    ------------
    allow_one_violation : bool, default=False
        Whether to allow violating one of the rules for a molecule. This makes the
        filter less restrictive.

    return_indicators : bool, default=False
        Whether to return a binary vector with indicators which molecules pass the
        filter, instead of list of molecules.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform_x_y` and
        :meth:`transform` are parallelized over the input molecules. ``None`` means 1
        unless in a :obj:`joblib.parallel_backend` context. ``-1`` means using all
        processors. See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when filtering molecules.

    References
    -----------
    .. [1] `Morelli, X., Bourgeas, R., & Roche, P.
        "Chemical and structural lessons from recent successes in protein–protein interaction inhibition (2P2I)."
        Current Opinion in Chemical Biology, 15(4), 475–481.
        <https://doi.org/10.1016/j.cbpa.2011.05.024>`_

    Examples
    ---------
    >>> from skfp.filters import RuleOfFourFilter
    >>> smiles = ['c1ccc2oc(-c3ccc(Nc4nc(N5CCCCC5)nc(N5CCOCC5)n4)cc3)nc2c1', \
    'c1nc(N2CCOCC2)c2sc3nc(N4CCOCC4)c4c(c3c2n1)CCCC4']
    >>> filt = RuleOfFourFilter()
    >>> filt
    RuleOfFourFilter()
    >>> filtered_mols = filt.transform(smiles)
    >>> filtered_mols
    ['c1ccc2oc(-c3ccc(Nc4nc(N5CCCCC5)nc(N5CCOCC5)n4)cc3)nc2c1']
    """

    def __init__(
        self,
        allow_one_violation: bool = False,
        return_indicators: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            allow_one_violation=allow_one_violation,
            return_indicators=return_indicators,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )

    def _apply_mol_filter(self, mol: Mol) -> bool:
        rules = [
            MolWt(mol) >= 400,
            MolLogP(mol) >= 4,
            CalcNumHBA(mol) >= 4,
            CalcNumRings(mol) >= 4,
        ]

        passed_rules = sum(rules)

        if self.allow_one_violation:
            return passed_rules >= len(rules) - 1
        else:
            return passed_rules == len(rules)



from collections.abc import Sequence
from numbers import Integral
from typing import Optional, Union

import numpy as np
from rdkit.Chem import Mol
from scipy.sparse import csr_array
from sklearn.utils._param_validation import Interval, InvalidParameterError, StrOptions

from skfp.bases import BaseFingerprintTransformer
from skfp.utils import ensure_mols, require_mols_with_conf_ids


class PharmacophoreFingerprint(BaseFingerprintTransformer):
    """
    Pharmacophore fingerprint.

    The implementation uses RDKit. This is a hashed fingerprint, where fragments are
    computed based on N-point tuples, using pharmacophoric points.

    An N-point pharmacophoric structure encodes N pharmacophoric points and pairwise
    distances between them, e.g. 3-point pharmacophore uses 6-element tuples
    (P1 D12 P2 D23 P3 D13). P is a pharmacophoric point, atom or subgraph, of a particular
    type (see below), and Dij is a topological distance (shortest path) between points
    i and j. Distance values are limited to 8 (higher values are capped at 8).

    Pharmacophoric point types are (based on SMARTS patterns definitions from [1]_):
    - hydrophobic atom
    - hydrogen bond donor
    - hydrogen bond acceptor
    - aromatic attachment
    - aliphatic attachment
    - "unusual" atom (not H, C, N, O, F, S, Cl, Br, I)
    - basic group
    - acidic group

    By default, 2-point and 3-point pharmacophores are used, resulting in 39972-element
    vector. Alternatively, folding can be applied. Note that due to RDKit limitations,
    only 2-point and 3-point variants are available.

    Note that this is by far the slowest fingerprint, particularly for larger molecules.
    This is due to the 3-point pharmacophore calculation. Consider filtering out large
    (heavy) molecules or setting ``max_points=2`` if it takes too long.

    Parameters
    ----------
    variant: {"raw_bits", "folded"} = "raw_bits"
        Whether to return raw bit values, or to fold them. Length of raw bits variant
        depends on used N-points, see ``n_features_out`` attribute.

    min_points: int, default=2
        Lower bound of N-point pharmacophore. Must be 2 or 3, and less or equal to
        ``max_points``.

    max_points: int, default=3
        Upper bound of N-point pharmacophore. Must be 2 or 3, and greater or equal to
        ``min_points``.

    fp_size : int, default=2048
        Size of output vectors, i.e. number of bits for each fingerprint. Only used
        for `"folded"` variant. Must be positive.

    use_3D : bool, default=False
        Whether to use 3D Euclidean distance matrix, instead of topological distance.
        Binning is used to discretize values into values 0-8.

    count : bool, default=False
        Whether to return binary (bit) features, or their counts.

    sparse : bool, default=False
        Whether to return dense NumPy array, or sparse SciPy CSR array.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized
        over the input molecules. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when computing fingerprints.

    Attributes
    ----------
    n_features_out : int, default=39972
        Number of output features, size of fingerprints. For ``"folded"`` variant,
        it is equal to ``fp_size``. For ``"raw_bits"`` variant, it depends on
        ``min_points`` and ``max_points``: 252 for (2,2), 39720 for (3,3), and
        39972 for (2,3).

    requires_conformers : bool
        Whether the fingerprint is 3D-based and requires molecules with conformers as
        inputs, with ``conf_id`` integer property set. This depends on the ``use_3D``
        parameter, and has the same value.

    References
    ----------
    .. [1] `A Gobbi, D Poppinger
        "Genetic optimization of combinatorial libraries"
        Biotechnology and Bioengineering: Volume 61, Issue 1, Winter 1998, Pages 47-54
        <https://doi.org/10.1002/(SICI)1097-0290(199824)61:1%3C47::AID-BIT9%3E3.0.CO;2-Z>`_

    Examples
    --------
    >>> from skfp.fingerprints import PharmacophoreFingerprint
    >>> smiles = ["O", "CC", "[C-]#N", "CC=O"]
    >>> fp = PharmacophoreFingerprint()
    >>> fp
    PharmacophoreFingerprint()

    >>> fp.transform(smiles)
    array([[0, 0, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 0, 0, 0],
           [0, 0, 0, ..., 0, 0, 0]], dtype=uint8)

    """

    _parameter_constraints: dict = {
        **BaseFingerprintTransformer._parameter_constraints,
        "variant": [StrOptions({"raw_bits", "folded"})],
        "min_points": [Interval(Integral, 2, 3, closed="both")],
        "max_points": [Interval(Integral, 2, 3, closed="both")],
        "fp_size": [Interval(Integral, 1, None, closed="left")],
        "use_3D": ["boolean"],
    }

    def __init__(
        self,
        variant: str = "raw_bits",
        min_points: int = 2,
        max_points: int = 3,
        fp_size: int = 2048,
        use_3D: bool = False,
        count: bool = False,
        sparse: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        n_features_out = self._get_n_features_out(
            variant, min_points, max_points, fp_size
        )
        super().__init__(
            n_features_out=n_features_out,
            requires_conformers=use_3D,
            count=count,
            sparse=sparse,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )
        self.variant = variant
        self.min_points = min_points
        self.max_points = max_points
        self.fp_size = fp_size
        self.use_3D = use_3D

    def _validate_params(self) -> None:
        super()._validate_params()
        if self.max_points < self.min_points:
            raise InvalidParameterError(
                f"The max_points parameter of {self.__class__.__name__} must be "
                f"greater or equal to min_points, got: "
                f"min_points={self.min_points}, max_points={self.max_points}"
            )

    def _get_n_features_out(
        self, variant: str, min_points: int, max_points: int, fp_size: int
    ) -> int:
        if variant == "folded":
            return fp_size

        # "raw_bits"
        if min_points == max_points == 2:
            return 252
        elif min_points == max_points == 3:
            return 39720
        elif min_points == 2 and max_points == 3:
            return 39972
        else:
            raise ValueError(
                "min_points and max_points must be 2 or 3, "
                "and min_points <= max_points, got:"
                f"min_points={min_points}, max_points={max_points}"
            )

    def _calculate_fingerprint(
        self, X: Sequence[Union[str, Mol]]
    ) -> Union[np.ndarray, csr_array]:
        from rdkit.Chem import Get3DDistanceMatrix
        from rdkit.Chem.ChemicalFeatures import BuildFeatureFactoryFromString
        from rdkit.Chem.Pharm2D import Gobbi_Pharm2D
        from rdkit.Chem.Pharm2D.Generate import Gen2DFingerprint
        from rdkit.Chem.Pharm2D.SigFactory import SigFactory

        atom_features = BuildFeatureFactoryFromString(Gobbi_Pharm2D.fdef)
        factory = SigFactory(
            atom_features,
            minPointCount=self.min_points,
            maxPointCount=self.max_points,
            useCounts=self.count,
        )
        factory.SetBins(Gobbi_Pharm2D.defaultBins)
        factory.Init()

        if not self.use_3D:
            X = ensure_mols(X)
            X = [Gen2DFingerprint(mol, factory) for mol in X]
        else:
            X = require_mols_with_conf_ids(X)
            X = [
                Gen2DFingerprint(
                    mol,
                    factory,
                    dMat=Get3DDistanceMatrix(mol, confId=mol.GetIntProp("conf_id")),
                )
                for mol in X
            ]

        if self.variant == "folded":
            # X at this point is a list of RDKit fingerprints, but MyPy doesn't get it
            return self._hash_fingerprint_bits(
                X,  # type: ignore
                fp_size=self.fp_size,
                count=self.count,
                sparse=self.sparse,
            )

        dtype = np.uint32 if self.count else np.uint8

        if self.sparse:
            return csr_array(X, dtype=dtype)
        else:
            return np.array(X, dtype=dtype)



from typing import Optional

from rdkit.Chem import GetFormalCharge, Mol
from rdkit.Chem.Crippen import MolLogP
from rdkit.Chem.Descriptors import MolWt
from rdkit.Chem.rdMolDescriptors import (
    CalcNumHBA,
    CalcNumHBD,
    CalcNumRings,
    CalcNumRotatableBonds,
    CalcTPSA,
)

from skfp.bases.base_filter import BaseFilter
from skfp.filters.utils import (
    get_max_ring_size,
    get_non_carbon_to_carbon_ratio,
    get_num_carbon_atoms,
    get_num_charged_functional_groups,
    get_num_rigid_bonds,
)


class ZINCDruglikeFilter(BaseFilter):
    """
    ZINC druglike filter.

    Designed to keep only drug-like molecules [1]_. Based only on physico-chemical
    properties [2]_, since SMARTS for additional rules are not publicly available.
    See "filter_light.txt" section in the supplementary material of the original
    paper [1]_ for details.

    Molecule must fulfill conditions:

    - molecular weight in range ``[60, 600]``
    - logP in range ``[-4, 6]``
    - HBA <= 11
    - HBD <= 6
    - TPSA <= 150
    - number of rotatable bonds <= 12
    - number of rigid bonds <= 50
    - number of rings <= 7
    - max ring size <= 12
    - number of carbobs >= 3
    - non-carbons to carbons ratio <= 2.0
    - number of charged functional groups <= 4
    - total formal charge in range ``[-4, 4]``

    Parameters
    ----------
    allow_one_violation : bool, default=False
        Whether to allow violating one of the rules for a molecule. This makes the
        filter less restrictive.

    return_indicators : bool, default=False
        Whether to return a binary vector with indicators which molecules pass the
        filter, instead of list of molecules.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform_x_y` and
        :meth:`transform` are parallelized over the input molecules. ``None`` means 1
        unless in a :obj:`joblib.parallel_backend` context. ``-1`` means using all
        processors. See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    verbose : int, default=0
        Controls the verbosity when filtering molecules.

    References
    ----------
    .. [1] `John J. Irwin, Brian K. Shoichet
        "ZINC − A Free Database of Commercially Available Compounds for Virtual Screening"
        J. Chem. Inf. Model. 2005, 45, 1, 177–182
        <https://doi.org/10.1021/ci049714+>`_

    .. [2] `Details of physico-chemical property filters available in FAF-Drugs4
        <https://fafdrugs4.rpbs.univ-paris-diderot.fr/filters.html>`_

    Examples
    --------
    >>> from skfp.filters import ZINCDruglikeFilter
    >>> smiles = ["C", "CC(=O)Nc1ccc(O)cc1"]
    >>> filt = ZINCDruglikeFilter()
    >>> filt
    ZINCDruglikeFilter()

    >>> filtered_mols = filt.transform(smiles)
    >>> filtered_mols
    ['CC(=O)Nc1ccc(O)cc1']
    """

    def __init__(
        self,
        allow_one_violation: bool = False,
        return_indicators: bool = False,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        verbose: int = 0,
    ):
        super().__init__(
            allow_one_violation=allow_one_violation,
            return_indicators=return_indicators,
            n_jobs=n_jobs,
            batch_size=batch_size,
            verbose=verbose,
        )

    def _apply_mol_filter(self, mol: Mol) -> bool:
        rules = [
            60 <= MolWt(mol) <= 600,
            -4 <= MolLogP(mol) <= 6,
            CalcNumHBA(mol) <= 11,
            CalcNumHBD(mol) <= 6,
            CalcTPSA(mol) <= 150,
            CalcNumRotatableBonds(mol) <= 12,
            get_num_rigid_bonds(mol) <= 50,
            CalcNumRings(mol) <= 7,
            get_max_ring_size(mol) <= 12,
            get_num_carbon_atoms(mol) >= 3,
            get_non_carbon_to_carbon_ratio(mol) <= 2,
            get_num_charged_functional_groups(mol) <= 4,
            -4 <= GetFormalCharge(mol) <= 4,
        ]
        passed_rules = sum(rules)

        if self.allow_one_violation:
            return passed_rules >= len(rules) - 1
        else:
            return passed_rules == len(rules)



from collections.abc import Sequence
from contextlib import nullcontext
from typing import Optional

from rdkit.Chem import Mol, MolFromFASTA
from sklearn.utils._param_validation import Options

from skfp.bases import BasePreprocessor
from skfp.utils import no_rdkit_logs
from skfp.utils.validators import check_strings


class MolFromAminoseqTransformer(BasePreprocessor):
    """
    Creates RDKit ``Mol`` objects from amino-acid sequence strings.

    Inputs are either sequences in FASTA format [1]_, or plain strings with
    amino-acid sequences.

    Parameters
    ----------
    sanitize : bool, default=True
        Whether to perform sanitization, i.e. basic validity checks, on created
        molecules. For details see RDKit documentation [2]_.

    flavor : int, default=0
        Type of molecule. See RDKit documentation [3]_ for more details.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`transform` is parallelized
        over the input molecules. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context. ``-1`` means using all processors.
        See Scikit-learn documentation on ``n_jobs`` for more details.

    batch_size : int, default=None
        Number of inputs processed in each batch. ``None`` divides input data into
        equal-sized parts, as many as ``n_jobs``.

    suppress_warnings: bool, default=False
        Whether to suppress warnings and errors on loading molecules.

    verbose : int, default=0
        Controls the verbosity when processing molecules.

    References
    ----------
    .. [1] `Lipman DJ, Pearson WR.
        "Rapid and sensitive protein similarity searches."
        Science. 1985 Mar 22; 227(4693):1435-41.
        <https://pubmed.ncbi.nlm.nih.gov/2983426/>`_

    .. [2] `Gregory Landrum
        "The RDKit Book: Molecular Sanitization"
        <https://www.rdkit.org/docs/RDKit_Book.html#molecular-sanitization>`_

    .. [3] `RDKit documentation - MolFromFASTA
        <https://www.rdkit.org/docs/source/rdkit.Chem.rdmolfiles.html#rdkit.Chem.rdmolfiles.MolFromFASTA>`_

    Examples
    --------
    >>> from skfp.preprocessing import MolFromAminoseqTransformer
    >>> sequences = ["KWLRRVWRWWR","FLPAIGRVLSGIL","ILGKLLSTAWGLLSKL",]
    >>> mol_from_aminoseq = MolFromAminoseqTransformer()
    >>> mol_from_aminoseq
    MolFromAminoseqTransformer()

    >>> mol_from_aminoseq.transform(sequences)  # doctest: +SKIP
        [<rdkit.Chem.rdchem.Mol>,
         <rdkit.Chem.rdchem.Mol>,
         <rdkit.Chem.rdchem.Mol>,
         <rdkit.Chem.rdchem.Mol>]
    """

    _parameter_constraints: dict = {
        **BasePreprocessor._parameter_constraints,
        "sanitize": ["boolean"],
        "flavor": [Options(int, {0, 1, 2, 3, 4, 5, 6, 7, 8, 9})],
    }

    def __init__(
        self,
        sanitize: bool = True,
        flavor: int = 0,
        n_jobs: Optional[int] = None,
        batch_size: Optional[int] = None,
        suppress_warnings: bool = False,
        verbose: int = 0,
    ):
        super().__init__(
            n_jobs=n_jobs,
            batch_size=batch_size,
            suppress_warnings=suppress_warnings,
            verbose=verbose,
        )
        self.sanitize = sanitize
        self.flavor = flavor

    def transform(self, X, copy: bool = False) -> list[Mol]:
        """
        Create RDKit ``Mol`` objects from amino-acid sequence strings. If ``valid_only``
        is set toTrue, returns only a subset of molecules which could be successfully
        loaded.

        Parameters
        ----------
        X : {sequence, array-like} of shape (n_samples,)
            Sequence containing amino-acid sequence strings.

        copy : bool, default=False
            Unused, kept for Scikit-learn compatibility.

        Returns
        -------
        X : list of shape (n_samples_conf_gen,)
            List with RDKit ``Mol`` objects.
        """
        return super().transform(X, copy)

    def _transform_batch(self, X: Sequence[str]) -> list[Mol]:
        with no_rdkit_logs() if self.suppress_warnings else nullcontext():
            check_strings(X)
            return [
                MolFromFASTA(fst, sanitize=self.sanitize, flavor=self.flavor)
                for fst in X
            ]



from contextlib import contextmanager

from rdkit.rdBase import BlockLogs


@contextmanager
def no_rdkit_logs():
    try:
        _rdkit_logs_blocker = BlockLogs()
        yield _rdkit_logs_blocker
    finally:
        del _rdkit_logs_blocker



import os
from typing import Optional, Union

import numpy as np
import pandas as pd
from sklearn.utils._param_validation import validate_params

from skfp.datasets.utils import fetch_dataset, get_mol_strings_and_labels


@validate_params(
    {
        "data_dir": [None, str, os.PathLike],
        "as_frame": ["boolean"],
        "verbose": ["boolean"],
    },
    prefer_skip_nested_validation=True,
)
def load_toxcast(
    data_dir: Optional[Union[str, os.PathLike]] = None,
    as_frame: bool = False,
    verbose: bool = False,
) -> Union[pd.DataFrame, tuple[list[str]], np.ndarray]:
    """
    Load and return the ToxCast dataset.

    The task is to predict 617 toxicity targets from a large library of compounds
    based on in vitro high-throughput screening. All tasks are binary.

    Note that targets have missing values. Algorithms should be evaluated only on
    present labels. For training data, you may want to impute them, e.g. with zeros.

    ==================   ========================
    Tasks                                     617
    Task type            multitask classification
    Total samples                            8576
    Recommended split                    scaffold
    Recommended metric                      AUROC
    ==================   ========================

    Parameters
    ----------
    data_dir : {None, str, path-like}, default=None
        Path to the root data directory. If ``None``, currently set scikit-learn directory
        is used, by default `$HOME/scikit_learn_data`.

    as_frame : bool, default=False
        If True, returns the raw DataFrame with columns "SMILES" and 617 label columns,
        with names corresponding to toxicity targets (see [1]_ and [2]_ for details).
        Otherwise, returns SMILES as list of strings, and labels as a NumPy array.
        Labels are 2D NumPy float array with binary labels and missing values.

    verbose : bool, default=False
        If True, progress bar will be shown for downloading or loading files.

    Returns
    -------
    data : pd.DataFrame or tuple(list[str], np.ndarray)
        Depending on the ``as_frame`` argument, one of:
        - Pandas DataFrame with columns "SMILES" and 617 label columns
        - tuple of: list of strings (SMILES), NumPy array (labels)

    References
    ----------
    .. [1] `Ann M. Richard et al.
        "ToxCast Chemical Landscape: Paving the Road to 21st Century Toxicology"
        Chem. Res. Toxicol. 2016, 29, 8, 1225–1251
        <https://pubs.acs.org/doi/10.1021/acs.chemrestox.6b00135>`_

    .. [2] `Zhenqin Wu et al.
        "MoleculeNet: a benchmark for molecular machine learning"
        Chem. Sci., 2018,9, 513-530
        <https://pubs.rsc.org/en/content/articlelanding/2018/sc/c7sc02664a>`_
    """
    df = fetch_dataset(
        data_dir,
        dataset_name="MoleculeNet_ToxCast",
        filename="toxcast.csv",
        verbose=verbose,
    )
    return df if as_frame else get_mol_strings_and_labels(df)



from collections.abc import Sequence
from itertools import chain
from typing import Any, Optional, Union

import numpy as np
from sklearn.model_selection._split import _validate_shuffle_split
from sklearn.utils import _safe_indexing


def ensure_nonempty_subset(data: list, subset: str) -> None:
    """
    Check if the provided list is empty.
    """
    if len(data) == 0:
        raise ValueError(f"{subset.capitalize()} subset is empty")


def split_additional_data(
    additional_data: list[Sequence[Any]], *indices_lists: list[int]
) -> list[Sequence[Any]]:
    """
    Split additional data based on indices lists.
    """
    return list(
        chain.from_iterable(
            (_safe_indexing(a, indices),)
            for a in additional_data
            for indices in indices_lists
        )
    )


def validate_train_test_split_sizes(
    train_size: Optional[Union[float, int]],
    test_size: Optional[Union[float, int]],
    n_samples: int,
) -> tuple[int, int]:
    """
    Ensure the sum of train_size and test_size equals 1.0 and provide default values
    if necessary. Returns integers with number of rows in those subsets.
    """
    return _validate_shuffle_split(
        n_samples, test_size, train_size, default_test_size=0.2
    )


def validate_train_valid_test_split_sizes(
    train_size: Optional[Union[float, int]],
    valid_size: Optional[Union[float, int]],
    test_size: Optional[Union[float, int]],
    n_samples: int,
) -> tuple[int, int, int]:
    """
    Ensure the sum of train_size, valid_size and test_size equals 1.0 and provide
    default values if necessary. Returns integers with number of rows in those subsets.
    """
    if train_size is None and valid_size is None and test_size is None:
        train_size = 0.8
        valid_size = 0.1
        test_size = 0.1

    sizes = (train_size, valid_size, test_size)
    if None in sizes:
        raise ValueError(f"All of the sizes must be provided, got: {sizes}")

    test_size_type = np.asarray(test_size).dtype.kind
    valid_size_type = np.asarray(valid_size).dtype.kind
    train_size_type = np.asarray(train_size).dtype.kind

    train_size: Union[int, float]
    valid_size: Union[int, float]
    test_size: Union[int, float]

    _check_subset_size(train_size, n_samples, "train")
    _check_subset_size(valid_size, n_samples, "valid")
    _check_subset_size(test_size, n_samples, "test")

    if (
        train_size_type == "f"
        and valid_size_type == "f"
        and test_size_type == "f"
        and train_size + valid_size + test_size > 1
    ):
        raise ValueError(
            f"The sum of float train_size, valid_size and test_size = "
            f"{train_size + valid_size + test_size}, should be in the (0, 1) "
            f"range."
        )

    n_test = int(test_size * n_samples) if test_size_type == "f" else test_size
    n_valid = int(valid_size * n_samples) if valid_size_type == "f" else valid_size
    n_train = n_samples - (n_test + n_valid) if valid_size_type == "f" else train_size

    if not n_train or not n_valid or not n_test:
        raise ValueError(
            f"With current sizes of train_size={train_size}, valid_size={valid_size}, "
            f"test_size={test_size}, and n_samples={n_samples}, one of the sets will "
            f"be empty."
        )
    if n_train + n_valid + n_test != n_samples:
        raise ValueError(
            f"The sum of train, valid and test sizes must be equal to the "
            f"n_samples={n_samples}, got: "
            f"n_train={n_train} (train_size={train_size}), "
            f"n_valid={n_valid} (valid_size={valid_size}), "
            f"n_test={n_test} (test_size={test_size})."
        )

    return int(n_train), int(n_valid), int(n_test)


def _check_subset_size(size: Union[float, int], n_samples: int, subset: str) -> None:
    size_type = np.asarray(size).dtype.kind
    subset_size_str = f"{subset}_size"  # e.g. "train_size", "test_size"

    if size_type not in ("i", "f"):
        raise ValueError(f"Invalid value for {subset_size_str}: {size}")

    if (
        size_type == "i"
        and (size >= n_samples or size <= 0)
        or size_type == "f"
        and (size <= 0 or size >= 1)
    ):
        raise ValueError(
            f"{subset_size_str}={size} should be either positive and smaller than "
            f"the number of samples {n_samples} or a float in the (0, 1) range"
        )
